from __future__ import annotations

import inspect
import pathlib
import time
import uuid
from datetime import timedelta

import pyarrow as pa
import pyarrow.compute as pc
import pytest
import toolz

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
    ParquetTTLSnapshotCache,
    SourceCache,
    SourceSnapshotCache,
)
from xorq.common.utils.inspect_utils import get_python_version_no_dot
from xorq.expr.udf import (
    agg,
)
from xorq.tests.util import (
    assert_frame_equal,
)
from xorq.vendor.ibis import _
from xorq.vendor.ibis.expr.types.relations import CACHED_NODE_NAME_PLACEHOLDER


KEY_PREFIX = xo.config.options.cache.key_prefix


@pytest.fixture
def cached_two(ls_con, batting, tmp_path):
    parquet_cache = ParquetCache.from_kwargs(source=ls_con, relative_path=tmp_path)
    return (
        batting[lambda t: t.yearID > 2014]
        .cache()[lambda t: t.stint == 1]
        .cache(cache=parquet_cache)
    )


@pytest.fixture(scope="function")
def pg_alltypes(pg):
    return pg.table("functional_alltypes")


def test_cache_simple(con, alltypes, alltypes_df):
    initial_tables = con.list_tables()

    expr = alltypes.select(
        alltypes.smallint_col, alltypes.int_col, alltypes.float_col
    ).filter(
        [
            alltypes.float_col > 0,
            alltypes.smallint_col == 9,
            alltypes.int_col < alltypes.float_col * 2,
        ]
    )
    cached = expr.cache(cache=SourceCache.from_kwargs(source=con))
    tables_after_caching = con.list_tables()

    expected = alltypes_df[
        (alltypes_df["float_col"] > 0)
        & (alltypes_df["smallint_col"] == 9)
        & (alltypes_df["int_col"] < alltypes_df["float_col"] * 2)
    ][["smallint_col", "int_col", "float_col"]]

    executed = cached.execute()
    tables_after_executing = con.list_tables()

    assert_frame_equal(executed, expected)
    assert not any(
        table_name.startswith(KEY_PREFIX)
        for table_name in set(tables_after_caching).difference(initial_tables)
    )
    assert any(
        table_name.startswith(KEY_PREFIX)
        for table_name in set(tables_after_executing).difference(initial_tables)
    )


def test_cache_multiple_times(con, alltypes, alltypes_df):
    expr = alltypes.select(
        alltypes.smallint_col, alltypes.int_col, alltypes.float_col
    ).filter(
        [
            alltypes.float_col > 0,
            alltypes.smallint_col == 9,
            alltypes.int_col < alltypes.float_col * 2,
        ]
    )
    cached = expr.cache()

    # reassign the expression
    expr = alltypes.select(
        alltypes.smallint_col, alltypes.int_col, alltypes.float_col
    ).filter(
        [
            alltypes.float_col > 0,
            alltypes.smallint_col == 9,
            alltypes.int_col < alltypes.float_col * 2,
        ]
    )

    re_cached = expr.cache()

    first = cached.execute()
    tables_after_first_caching = con.list_tables()

    second = re_cached.execute()
    tables_after_second_caching = con.list_tables()

    expected = alltypes_df[
        (alltypes_df["float_col"] > 0)
        & (alltypes_df["smallint_col"] == 9)
        & (alltypes_df["int_col"] < alltypes_df["float_col"] * 2)
    ][["smallint_col", "int_col", "float_col"]]

    assert_frame_equal(first, expected)
    assert_frame_equal(second, expected)

    first_tables = [t for t in tables_after_first_caching if t.startswith(KEY_PREFIX)]
    second_tables = [t for t in tables_after_second_caching if t.startswith(KEY_PREFIX)]

    assert sorted(first_tables) == sorted(second_tables)


def test_cache_to_sql(alltypes):
    expr = alltypes.select(
        alltypes.smallint_col, alltypes.int_col, alltypes.float_col
    ).filter(
        [
            alltypes.float_col > 0,
            alltypes.smallint_col == 9,
            alltypes.int_col < alltypes.float_col * 2,
        ]
    )
    cached = expr.cache()

    assert CACHED_NODE_NAME_PLACEHOLDER in xo.to_sql(cached)


def test_op_after_cache(alltypes):
    expr = alltypes.select(alltypes.smallint_col, alltypes.int_col, alltypes.float_col)
    cached = expr.cache()
    cached = cached.filter(
        [
            _.float_col > 0,
            _.smallint_col == 9,
            _.int_col < _.float_col * 2,
        ]
    )

    full_expr = expr.filter(
        [
            alltypes.float_col > 0,
            alltypes.smallint_col == 9,
            alltypes.int_col < alltypes.float_col * 2,
        ]
    )

    actual = cached.execute()
    expected = full_expr.execute()

    assert_frame_equal(actual, expected)

    assert CACHED_NODE_NAME_PLACEHOLDER in xo.to_sql(cached)


def test_cache_recreate(alltypes):
    def make_expr(alltypes):
        return alltypes.select(
            alltypes.smallint_col, alltypes.int_col, alltypes.float_col
        ).filter(
            [
                alltypes.float_col > 0,
                alltypes.smallint_col == 9,
                alltypes.int_col < alltypes.float_col * 2,
            ]
        )

    alltypes_df = alltypes.execute()
    cons = (con0, con1) = xo.connect(), xo.connect()
    ts = tuple(con.create_table("alltypes", alltypes_df) for con in cons)
    exprs = tuple(make_expr(t) for t in ts)

    for con, expr in zip(cons, exprs):
        # FIXME: execute one, simply check the other returns true for `expr.ls.exists()`
        expr.cache(cache=SourceCache.from_kwargs(source=con)).execute()

    (con_cached_tables0, con_cached_tables1) = (
        {
            table_name
            for table_name in con.list_tables()
            if table_name.startswith(KEY_PREFIX)
        }
        for con in cons
    )

    assert con_cached_tables0
    assert con_cached_tables0 == con_cached_tables1
    for table_name in con_cached_tables1:
        assert_frame_equal(
            con0.table(table_name).to_pandas(), con1.table(table_name).to_pandas()
        )


def test_cache_execution(alltypes):
    cached = (
        alltypes.select(alltypes.smallint_col, alltypes.int_col, alltypes.float_col)
        .cache()
        .filter(
            [
                _.float_col > 0,
                _.smallint_col == 9,
                _.int_col < _.float_col * 2,
            ]
        )
        .select(_.int_col * 4)
        .cache()
    )

    actual = cached.execute()

    expected = (
        alltypes.select(alltypes.smallint_col, alltypes.int_col, alltypes.float_col)
        .filter(
            [
                alltypes.float_col > 0,
                alltypes.smallint_col == 9,
                alltypes.int_col < alltypes.float_col * 2,
            ]
        )
        .select(alltypes.int_col * 4)
        .execute()
    )

    assert_frame_equal(actual, expected)


def test_parquet_cache_storage(tmp_path, alltypes_df):
    tmp_path = pathlib.Path(tmp_path)
    path = tmp_path.joinpath("to-delete.parquet")

    con = xo.connect()
    alltypes_df.to_parquet(path)
    t = con.read_parquet(path, "t")
    cols = ["id", "bool_col", "float_col", "string_col"]
    expr = t[cols]
    expected = alltypes_df[cols]
    source = expr._find_backend()
    cache = ParquetCache.from_kwargs(
        source=source,
        relative_path=tmp_path.joinpath("parquet-cache-storage"),
    )
    cached = expr.cache(cache=cache)
    actual = cached.execute()
    assert_frame_equal(actual, expected)

    # the file must exist and have the same schema
    alltypes_df.head(1).to_parquet(path)
    actual = cached.execute()
    assert_frame_equal(actual, expected)

    path.unlink()
    pattern = "No data to cache"
    with pytest.raises(ValueError, match=pattern):
        # if the file doesn't exist, we get a failure, even for cached
        cached.execute()


def test_parquet_remote_to_local(con, alltypes, tmp_path):
    tmp_path = pathlib.Path(tmp_path)

    expr = alltypes.select(
        alltypes.smallint_col, alltypes.int_col, alltypes.float_col
    ).filter(
        [
            alltypes.float_col > 0,
            alltypes.smallint_col == 9,
            alltypes.int_col < alltypes.float_col * 2,
        ]
    )
    cache = ParquetCache.from_kwargs(
        source=con,
        relative_path=tmp_path.joinpath("parquet-cache-cache"),
    )
    cached = expr.cache(cache=cache)
    expected = expr.execute()
    actual = cached.execute()
    assert_frame_equal(actual, expected)


def test_read_parquet_and_cache(con, parquet_dir, tmp_path):
    batting_path = parquet_dir / "batting.parquet"
    t = con.read_parquet(batting_path, table_name=f"parquet_batting-{uuid.uuid4()}")
    expr = t.cache(cache=ParquetCache.from_kwargs(source=con, relative_path=tmp_path))
    assert expr.execute() is not None


def test_read_parquet_and_cache_xorq(ls_con, parquet_dir, tmp_path):
    batting_path = parquet_dir / "batting.parquet"
    t = ls_con.read_parquet(batting_path, table_name=f"parquet_batting-{uuid.uuid4()}")
    expr = t.cache(
        cache=ParquetCache.from_kwargs(source=ls_con, relative_path=tmp_path)
    )
    assert expr.execute() is not None


def test_read_parquet_compute_and_cache(con, parquet_dir, tmp_path):
    batting_path = parquet_dir / "batting.parquet"
    t = con.read_parquet(batting_path, table_name=f"parquet_batting-{uuid.uuid4()}")
    expr = (
        t[t.yearID == 2015]
        .cache(cache=ParquetCache.from_kwargs(source=con, relative_path=tmp_path))
        .cache()
    )
    assert expr.execute() is not None


def test_read_csv_and_cache(ls_con, csv_dir, tmp_path):
    batting_path = csv_dir / "batting.csv"
    t = ls_con.read_csv(batting_path, table_name=f"csv_batting-{uuid.uuid4()}")
    expr = t.cache(
        cache=ParquetCache.from_kwargs(source=ls_con, relative_path=tmp_path)
    )
    assert expr.execute() is not None


def test_read_csv_compute_and_cache(ls_con, csv_dir, tmp_path):
    batting_path = csv_dir / "batting.csv"
    t = ls_con.read_csv(
        batting_path,
        table_name=f"csv_batting-{uuid.uuid4()}",
        schema_infer_max_records=50_000,
    )
    expr = (
        t[t.yearID == 2015]
        .cache(cache=ParquetCache.from_kwargs(source=ls_con, relative_path=tmp_path))
        .cache()
    )
    assert expr.execute() is not None


def test_repeated_cache(con, ls_con, tmp_path):
    cache = ParquetCache.from_kwargs(
        source=ls_con,
        relative_path=tmp_path,
    )
    t = (
        con.table("batting")[lambda t: t.yearID > 2014]
        .cache(cache=cache)[lambda t: t.stint == 1]
        .cache(cache=cache)
    )

    actual = t.execute()
    expected = con.table("batting").filter([_.yearID > 2014, _.stint == 1]).execute()

    assert_frame_equal(actual, expected)


def test_cache_default_path_set(batting, ls_con, tmp_path):
    xo.options.cache.default_relative_path = tmp_path

    cache = ParquetCache.from_kwargs(
        source=ls_con,
    )

    expr = batting[lambda t: t.yearID > 2014].limit(1).cache(cache=cache)

    result = expr.execute()

    cache_files = [
        path
        for path in tmp_path.iterdir()
        if path.is_file()
        and path.name.startswith(KEY_PREFIX)
        and path.name.endswith(".parquet")
    ]

    assert result is not None
    assert cache_files


@pytest.mark.snapshot_check
@pytest.mark.xfail
def test_udf_caching(ls_con, alltypes_df, snapshot):
    @xo.udf.scalar.pyarrow
    def my_mul(tinyint_col: dt.int16, smallint_col: dt.int16) -> dt.int16:
        return pc.multiply(tinyint_col, smallint_col)

    @toolz.curry
    def wrapper(f, t):
        # here: map pandas into pyarrow
        # goal: map pyarrow into pandas: so users can define functions in pandas land
        inner = f.__wrapped__
        kwargs = {k: pa.array(v) for k, v in t.items()}
        return inner(**kwargs)

    cols = list(inspect.signature(my_mul).parameters)

    expr = (
        ls_con.create_table("alltypes", alltypes_df)[cols]
        .pipe(lambda t: t.mutate(mulled=my_mul(*(t[col] for col in cols))))
        .cache()
    )
    from_ls = expr.execute()
    from_pandas = alltypes_df[cols].assign(mulled=wrapper(my_mul))
    assert from_ls.equals(from_pandas)

    py_version = f"py{get_python_version_no_dot()}"
    snapshot.assert_match(expr.ls.get_key(), f"{py_version}_udf_caching.txt")


@pytest.mark.snapshot_check
@pytest.mark.xfail
def test_udaf_caching(ls_con, alltypes_df, snapshot):
    def my_mul_sum(df):
        return df.sum().sum()

    cols = ["tinyint_col", "smallint_col"]
    ibis_output_type = dt.infer(alltypes_df[cols].sum().sum())
    by = "bool_col"
    name = "my_mul_sum"

    t = ls_con.create_table("alltypes", alltypes_df)
    agg_udf = agg.pandas_df(
        my_mul_sum,
        t[cols].schema(),
        ibis_output_type,
        name=name,
    )
    from_pandas = (
        alltypes_df.groupby(by)[cols]
        .apply(my_mul_sum)
        .rename(name)
        .reset_index()
        .sort_values(by)
    )
    expr = (
        t.group_by(by)
        .agg(**{name: agg_udf(*(t[col] for col in cols))})
        .order_by(by)
        .cache()
    )
    on_expr = t.group_by(by).agg(**{name: agg_udf.on_expr(t)}).order_by(by).cache()
    assert not expr.ls.exists()
    assert not on_expr.ls.exists()

    from_ls = expr.execute().sort_values(by="bool_col")
    assert_frame_equal(from_ls, from_pandas.sort_values(by="bool_col"))
    assert_frame_equal(from_ls, on_expr.execute().sort_values(by="bool_col"))
    assert expr.ls.exists()
    assert on_expr.ls.exists()

    py_version = f"py{get_python_version_no_dot()}"
    snapshot.assert_match(expr.ls.get_key(), f"{py_version}_test_udaf_caching.txt")
    snapshot.assert_match(on_expr.ls.get_key(), f"{py_version}_test_udaf_caching.txt")


def test_cache_exists_doesnt_materialize(cached_two):
    cache = cached_two.ls.cache
    assert not cache.exists(cached_two)
    assert not cache.exists(cached_two)
    assert not cached_two.ls.exists()
    cached_two.count().execute()
    assert cached_two.ls.exists()
    assert cache.exists(cached_two)


def test_ls_exists_doesnt_materialize(cached_two):
    cache = cached_two.ls.cache
    assert not cached_two.ls.exists()
    assert not cached_two.ls.exists()
    assert not cache.exists(cached_two)
    cached_two.count().execute()
    assert cached_two.ls.exists()
    assert cache.exists(cached_two)


@pytest.mark.parametrize(
    "cls",
    [ParquetSnapshotCache, ParquetCache, SourceSnapshotCache, SourceCache],
)
def test_cache_find_embedded_backend(cls, parquet_dir):
    con = xo.connect()
    astronauts_path = parquet_dir / "astronauts.parquet"
    cache = cls.from_kwargs(source=con)
    expr = con.read_parquet(astronauts_path).cache(cache=cache)
    assert expr._find_backend()._profile == con._profile


def test_parquet_ttl_snapshot_cache(ls_con, batting, tmp_path):
    n_rows = 1
    seconds = 1
    tmp_path = pathlib.Path(tmp_path)
    all_rows_path = tmp_path.joinpath("all_rows.parquet")
    n_rows_path = tmp_path.joinpath("n_rows.parquet")
    path = tmp_path.joinpath("data.parquet")
    expr = batting[lambda t: t.yearID > 2014][lambda t: t.stint == 1]
    expr.to_parquet(all_rows_path)
    assert xo.read_parquet(all_rows_path).count().execute() != n_rows
    expr.limit(n_rows).to_parquet(n_rows_path)
    path.write_bytes(all_rows_path.read_bytes())

    cache = ParquetTTLSnapshotCache.from_kwargs(
        source=ls_con, relative_path=tmp_path, ttl=timedelta(seconds=1)
    )
    expr = xo.deferred_read_parquet(
        path=path,
        con=ls_con,
    ).cache(cache=cache)
    assert not expr.ls.exists()
    m = expr.count().execute()
    assert expr.ls.exists()
    path.write_bytes(n_rows_path.read_bytes())
    assert expr.ls.exists()
    assert expr.count().execute() == m
    assert ls_con.read_parquet(path).count().execute() == n_rows
    time.sleep(seconds)
    assert not expr.ls.exists()
    assert expr.count().execute() == n_rows



"""GizmoSQL backend tests — single-backend and multi-engine."""

from __future__ import annotations

from operator import methodcaller

import pandas as pd
import pyarrow as pa
import pytest

import xorq.api as xo
from xorq.caching import ParquetCache, SourceCache


pytestmark = pytest.mark.gizmosql


# ── Single-backend tests ────────────────────────────────────────────────────


def test_connect(con):
    """Verify that the GizmoSQL connection is established."""
    assert con is not None
    assert con.name == "gizmosql"


def test_version(con):
    """Verify we can retrieve the server version."""
    version = con.version
    assert isinstance(version, str)
    assert len(version) > 0


def test_list_tables(con):
    """Verify that loaded test tables are listed."""
    tables = con.list_tables()
    assert isinstance(tables, list)
    assert "functional_alltypes" in tables
    assert "batting" in tables


def test_list_catalogs(con):
    """Verify that catalogs can be listed."""
    catalogs = con.list_catalogs()
    assert isinstance(catalogs, list)
    assert len(catalogs) > 0


def test_list_databases(con):
    """Verify that databases (schemas) can be listed."""
    databases = con.list_databases()
    assert isinstance(databases, list)
    assert "main" in databases


def test_current_catalog(con):
    """Verify current_catalog property works."""
    catalog = con.current_catalog
    assert isinstance(catalog, str)
    assert len(catalog) > 0


def test_current_database(con):
    """Verify current_database property works."""
    database = con.current_database
    assert isinstance(database, str)
    assert database == "main"


def test_execute(con, alltypes):
    """Verify executing an expression returns a pandas DataFrame."""
    result = alltypes.limit(10).execute()
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 10


def test_to_pyarrow(con, alltypes):
    """Verify to_pyarrow returns a PyArrow Table."""
    result = con.to_pyarrow(alltypes.limit(10))
    assert isinstance(result, pa.Table)
    assert result.num_rows == 10


def test_to_pyarrow_batches(con, alltypes):
    """Verify to_pyarrow_batches returns a RecordBatchReader."""
    reader = con.to_pyarrow_batches(alltypes.limit(10))
    assert isinstance(reader, pa.ipc.RecordBatchReader)
    table = reader.read_all()
    assert table.num_rows == 10


def test_create_and_drop_table(con, temp_table):
    """Verify table creation and deletion via DDL."""
    t = xo.memtable({"a": [1, 2, 3], "b": ["x", "y", "z"]})
    con.create_table(temp_table, t)

    assert temp_table in con.list_tables()

    result = con.table(temp_table).execute()
    assert len(result) == 3

    with con._safe_raw_sql(f'DROP TABLE IF EXISTS "{temp_table}"'):
        pass
    assert temp_table not in con.list_tables()


def test_read_parquet(con):
    """Verify read_parquet via _read_local_and_ingest."""
    from pathlib import Path

    root = Path(__file__).resolve().parents[5]
    parquet_path = root / "ci" / "ibis-testing-data" / "parquet" / "diamonds.parquet"
    if not parquet_path.exists():
        pytest.skip("Test data not available")

    t = con.read_parquet(parquet_path, table_name="test_diamonds_read")
    result = t.limit(5).execute()
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 5

    # Clean up
    with con._safe_raw_sql('DROP TABLE IF EXISTS "test_diamonds_read"'):
        pass


def test_memtable(con):
    """Verify in-memory table registration via ADBC."""
    t = xo.memtable({"x": [10, 20, 30], "y": ["a", "b", "c"]})
    result = con.execute(t)
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 3
    assert list(result["x"]) == [10, 20, 30]


def test_schema(con, alltypes):
    """Verify we can retrieve the schema of a table."""
    schema = alltypes.schema()
    assert len(schema) > 0
    assert "id" in schema


# ── into_backend tests ──────────────────────────────────────────────────────


@pytest.mark.parametrize("method", ["execute", "to_pyarrow", "to_pyarrow_batches"])
def test_into_backend_to_duckdb(con, batting, method):
    """GizmoSQL → DuckDB via into_backend, all output methods."""
    ddb_con = xo.duckdb.connect()
    expr = batting.limit(50).into_backend(ddb_con, "gz_batting")
    res = methodcaller(method)(expr)

    if isinstance(res, pa.RecordBatchReader):
        res = res.read_all()
    if isinstance(res, pa.Table):
        assert res.num_rows == 50
    else:
        assert len(res) == 50


@pytest.mark.parametrize("method", ["execute", "to_pyarrow", "to_pyarrow_batches"])
def test_into_backend_to_datafusion(con, batting, method):
    """GizmoSQL → DataFusion (xorq default) via into_backend."""
    xorq_con = xo.connect()
    expr = batting.limit(50).into_backend(xorq_con, "gz_batting")
    res = methodcaller(method)(expr)

    if isinstance(res, pa.RecordBatchReader):
        res = res.read_all()
    if isinstance(res, pa.Table):
        assert res.num_rows == 50
    else:
        assert len(res) == 50


def test_into_backend_join_on_duckdb(con, batting, awards_players):
    """Transfer two GizmoSQL tables to DuckDB, join there."""
    ddb_con = xo.duckdb.connect()

    batting_ddb = batting.filter(batting.yearID >= 2010).into_backend(
        ddb_con, "gz_batting"
    )
    awards_ddb = awards_players.into_backend(ddb_con, "gz_awards")

    expr = (
        batting_ddb.join(awards_ddb, predicates=["playerID", "yearID"])
        .limit(20)
        .select("playerID", "yearID", "awardID", "H")
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 20


def test_into_backend_double_hop(con, batting):
    """GizmoSQL → DuckDB → DataFusion (two hops)."""
    from xorq.vendor.ibis import _

    ddb_con = xo.duckdb.connect()
    xorq_con = xo.connect()

    expr = (
        batting.filter(batting.yearID == 2015)
        .select("playerID", "yearID", "teamID", "H")
        .into_backend(ddb_con, "gz_hop1")
        .filter(_.H > 100)
        .into_backend(xorq_con, "gz_hop2")
        .limit(10)
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 10


def test_into_backend_reverse_duckdb_to_gizmosql(con, batting):
    """DuckDB → GizmoSQL: push local data into the GizmoSQL server."""
    ddb_con = xo.duckdb.connect()
    local_table = ddb_con.create_table(
        "local_players",
        batting.limit(100).to_pyarrow(),
    )
    expr = local_table.into_backend(con, "ddb_to_gz")

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert len(res) == 100


# ── Caching tests ───────────────────────────────────────────────────────────


def test_into_backend_source_cache(con, batting):
    """GizmoSQL → DataFusion with SourceCache."""
    xorq_con = xo.connect()

    expr = (
        batting.filter(batting.yearID >= 2014)
        .limit(50)
        .into_backend(xorq_con, "gz_cached_batting")
        .cache(SourceCache.from_kwargs(source=xorq_con))
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 50


def test_into_backend_parquet_cache(con, batting, tmp_path):
    """GizmoSQL → DuckDB with ParquetCache."""
    ddb_con = xo.duckdb.connect()

    expr = (
        batting.filter(batting.yearID >= 2014)
        .limit(50)
        .into_backend(ddb_con, "gz_pq_batting")
        .cache(ParquetCache.from_kwargs(source=ddb_con, relative_path=tmp_path))
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 50

    # second execution should hit cache
    res2 = expr.execute()
    assert len(res2) == len(res)


def test_into_backend_chained_caches(con, batting, tmp_path):
    """GizmoSQL → DataFusion [SourceCache] → DuckDB [ParquetCache]."""
    xorq_con = xo.connect()
    ddb_con = xo.duckdb.connect()

    expr = (
        batting.filter(batting.yearID >= 2014)
        .limit(50)
        .into_backend(xorq_con, "gz_chain_batting")
        .cache(SourceCache.from_kwargs(source=xorq_con))
        .into_backend(ddb_con)
        .select("playerID", "yearID", "teamID", "H")
        .cache(ParquetCache.from_kwargs(source=ddb_con, relative_path=tmp_path))
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 50



from __future__ import annotations

import pickle

import numpy as np
import pandas as pd
import pytest

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq.caching import ParquetCache
from xorq.common.utils.defer_utils import (
    deferred_read_parquet,
)
from xorq.expr.ml.fit_lib import (
    DeferredFitOther,
    kv_encode_output,
)
from xorq.expr.ml.structer import KV_ENCODED_TYPE
from xorq.ml import (
    deferred_fit_predict_sklearn,
    deferred_fit_transform_series_sklearn,
)


sk_linear_model = pytest.importorskip("sklearn.linear_model")
sk_preprocessing = pytest.importorskip("sklearn.preprocessing")
sk_feature_selection = pytest.importorskip("sklearn.feature_selection")
sk_feature_extraction_text = pytest.importorskip("sklearn.feature_extraction.text")


def make_data():
    X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
    # y = 1 * x_0 + 2 * x_1 + 3
    y = np.dot(X, np.array([1, 2])) + 3
    df = pd.DataFrame(np.hstack((X, y[:, np.newaxis]))).rename(
        columns=lambda x: chr(x + ord("a"))
    )
    (*features, target) = df.columns
    return (df, features, target)


deferred_linear_regression = deferred_fit_predict_sklearn(
    cls=sk_linear_model.LinearRegression, return_type=dt.float64
)


def test_deferred_fit_predict_linear_regression(tmp_path):
    con = xo.connect()
    (df, features, target) = make_data()
    t = con.register(df, "t")

    # uncached run
    instance = deferred_linear_regression(t, target, features)
    (computed_kwargs_expr, predict_expr_udf) = (
        instance.deferred_model,
        instance.deferred_other,
    )
    model = computed_kwargs_expr.execute()
    predicted = t.mutate(predict_expr_udf.on_expr(t)).execute()

    # cached run
    cache = ParquetCache.from_kwargs(relative_path=tmp_path, source=con)
    instance = deferred_linear_regression(t, target, features, cache=cache)
    (computed_kwargs_expr, predict_expr_udf) = (
        instance.deferred_model,
        instance.deferred_other,
    )
    ((cached_model,),) = computed_kwargs_expr.execute().values
    cached_predicted = t.mutate(predict_expr_udf.on_expr(t)).execute()

    assert predicted.equals(cached_predicted)
    np.testing.assert_almost_equal(
        pickle.loads(model).coef_, pickle.loads(cached_model).coef_
    )


def test_deferred_fit_predict_linear_regression_multi_into_backend():
    con = xo.connect()
    (df, features, target) = make_data()
    t = (
        con.register(df, "t")
        .into_backend(xo.connect())[lambda t: t[features[0]] > 0]
        .into_backend(xo.connect())
    )

    instance = deferred_linear_regression(t, target, features)
    predict_expr_udf = instance.deferred_other
    predicted = t.mutate(predict_expr_udf.on_expr(t)).execute()
    assert not predicted.empty


def test_deferred_fit_transform_series_sklearn():
    cls = sk_feature_extraction_text.TfidfVectorizer
    transform_key = "transformed"
    col = "title"
    deferred_fit_transform_tfidf = deferred_fit_transform_series_sklearn(
        col=col, cls=cls, return_type=dt.Array(dt.float64)
    )
    con = xo.connect()
    train, test = deferred_read_parquet(
        xo.options.pins.get_path("hn-data-small.parquet"),
        con,
        "fetcher-input",
    ).pipe(
        xo.train_test_splits,
        unique_key="id",
        test_sizes=(0.9, 0.1),
    )
    instance = deferred_fit_transform_tfidf(train)
    deferred_transform = instance.deferred_other

    from_sklearn = test.execute().assign(
        **{
            transform_key: lambda t: (
                cls().fit(train.execute()[col]).transform(t[col]).toarray().tolist()
            )
        }
    )
    from_xo = test.mutate(**{transform_key: deferred_transform.on_expr}).execute()
    actual = from_xo[transform_key].apply(pd.Series)
    expected = from_sklearn[transform_key].apply(pd.Series)
    assert actual.equals(expected)


class TestDeferredFitOtherFromFittedStep:
    """Tests for DeferredFitOther.from_fitted_step factory method."""

    def test_from_fitted_step_transform_known_schema(self):
        """Test from_fitted_step returns struct deferred for StandardScaler."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0]})
        step = xo.Step.from_instance_name(
            sk_preprocessing.StandardScaler(), name="scaler"
        )
        fitted = step.fit(t, features=("a", "b"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred is not None
        assert deferred.expr is fitted.expr
        assert deferred.features == fitted.features
        # Known schema transformers should have "transformed" in name
        assert "transformed" in deferred.name_infix

    def test_from_fitted_step_transform_kv_encoded(self):
        """Test from_fitted_step returns encoded deferred for OneHotEncoder."""
        t = xo.memtable({"cat": ["a", "b", "c"]})
        step = xo.Step.from_instance_name(sk_preprocessing.OneHotEncoder(), name="ohe")
        fitted = step.fit(t, features=("cat",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred is not None
        assert deferred.expr is fitted.expr
        assert deferred.features == fitted.features
        # KV-encoded transformers should have encoded in name
        assert "encoded" in deferred.name_infix

    def test_from_fitted_step_transform_series_kv_encoded(self):
        """Test from_fitted_step returns series encoded deferred for TfidfVectorizer."""
        t = xo.memtable({"text": ["hello world", "foo bar"]})
        step = xo.Step.from_instance_name(
            sk_feature_extraction_text.TfidfVectorizer(), name="tfidf"
        )
        fitted = step.fit(t, features=("text",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred is not None
        assert deferred.expr is fitted.expr
        # TfidfVectorizer is series + KV-encoded
        assert "encoded" in deferred.name_infix

    def test_from_fitted_step_transform_with_target(self):
        """Test from_fitted_step passes target for supervised transformers."""
        t = xo.memtable(
            {"a": [1.0, 2.0, 3.0, 4.0], "b": [4.0, 3.0, 2.0, 1.0], "y": [0, 0, 1, 1]}
        )
        step = xo.Step.from_instance_name(
            sk_feature_selection.SelectKBest(k=1), name="skb"
        )
        fitted = step.fit(t, features=("a", "b"), target="y")

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred is not None
        assert deferred.target == "y"

    def test_from_fitted_step_predict(self):
        """Test from_fitted_step returns predict deferred for predictor."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0], "y": [0.0, 1.0]})
        step = xo.Step.from_instance_name(sk_linear_model.LinearRegression(), name="lr")
        fitted = step.fit(t, features=("a", "b"), target="y")

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred is not None
        assert deferred.expr is fitted.expr
        assert deferred.features == fitted.features
        assert deferred.target == "y"
        assert "predict" in deferred.name_infix

    def test_from_fitted_step_transform_with_params_executes(self):
        """Test from_fitted_step passes params correctly and executes."""
        t = xo.memtable({"cat": ["a", "b", "c"]})
        step = xo.Step.from_instance_name(
            sk_preprocessing.OneHotEncoder(sparse_output=False), name="ohe"
        )
        fitted = step.fit(t, features=("cat",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred is not None
        # Verify deferred can be used to transform
        col = deferred.deferred_other.on_expr(t)
        result = t.mutate(col).execute()
        # The column name is based on the UDF, check it exists and has KV-encoded data
        assert len(result.columns) == 2  # cat + transformed column
        transformed_col = [c for c in result.columns if c != "cat"][0]
        # Should have KV-encoded data
        first_row = result[transformed_col].iloc[0]
        assert all("key" in d and "value" in d for d in first_row)


sk_compose = pytest.importorskip("sklearn.compose")
sk_pipeline = pytest.importorskip("sklearn.pipeline")


class TestFromFittedStepMatchStatementBranches:
    """Tests verifying each branch of the match statement in from_fitted_step."""

    def test_series_kv_encoded_branch_name_infix(self):
        """Test (True, True, _) branch uses 'transformed_encoded' name_infix for TfidfVectorizer."""
        t = xo.memtable({"text": ["hello world", "foo bar"]})
        step = xo.Step.from_instance_name(
            sk_feature_extraction_text.TfidfVectorizer(), name="tfidf"
        )
        fitted = step.fit(t, features=("text",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred.name_infix == "transformed_encoded"
        assert deferred.return_type == KV_ENCODED_TYPE

    def test_pure_kv_encoded_branch_name_infix(self):
        """Test (False, True, _) branch uses 'transformed_encoded' name_infix for OneHotEncoder."""
        t = xo.memtable({"cat": ["a", "b", "c"]})
        step = xo.Step.from_instance_name(sk_preprocessing.OneHotEncoder(), name="ohe")
        fitted = step.fit(t, features=("cat",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred.name_infix == "transformed_encoded"
        assert deferred.return_type == KV_ENCODED_TYPE

    def test_struct_with_kv_fields_branch_name_infix(self):
        """Test (False, False, True) branch uses 'transformed_struct_kv' for ColumnTransformer with mixed output."""
        t = xo.memtable({"num": [1.0, 2.0, 3.0], "cat": ["a", "b", "c"]})
        ct = sk_compose.ColumnTransformer(
            [
                ("scaler", sk_preprocessing.StandardScaler(), ["num"]),
                ("encoder", sk_preprocessing.OneHotEncoder(), ["cat"]),
            ]
        )
        step = xo.Step.from_instance_name(ct, name="ct")
        fitted = step.fit(t, features=("num", "cat"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred.name_infix == "transformed_struct_kv"
        # Return type should be a struct containing both regular and KV-encoded fields
        assert deferred.return_type is not None

    def test_pure_struct_branch_name_infix(self):
        """Test default branch uses 'transformed' name_infix for StandardScaler."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0]})
        step = xo.Step.from_instance_name(
            sk_preprocessing.StandardScaler(), name="scaler"
        )
        fitted = step.fit(t, features=("a", "b"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred.name_infix == "transformed"
        # Return type should be a struct
        assert isinstance(deferred.return_type, dt.Struct)

    def test_predict_branch_name_infix(self):
        """Test predict branch uses 'predict' name_infix for LinearRegression."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0], "y": [0.0, 1.0]})
        step = xo.Step.from_instance_name(sk_linear_model.LinearRegression(), name="lr")
        fitted = step.fit(t, features=("a", "b"), target="y")

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred.name_infix == "predict"


class TestFromFittedStepFitFunctions:
    """Tests verifying correct fit function is used for each transformer type."""

    def test_series_transformer_uses_fit_sklearn_series(self):
        """Test TfidfVectorizer uses fit_sklearn_series with correct col parameter."""
        t = xo.memtable({"text": ["hello world", "foo bar"]})
        step = xo.Step.from_instance_name(
            sk_feature_extraction_text.TfidfVectorizer(), name="tfidf"
        )
        fitted = step.fit(t, features=("text",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # fit_sklearn_series is a curried function, verify it has the col parameter
        assert deferred.fit.keywords.get("col") == "text"

    def test_multi_column_transformer_uses_fit_sklearn_args(self):
        """Test StandardScaler with multiple columns uses fit_sklearn_args."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0]})
        step = xo.Step.from_instance_name(
            sk_preprocessing.StandardScaler(), name="scaler"
        )
        fitted = step.fit(t, features=("a", "b"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # fit_sklearn_args uses cls and params keywords
        assert "cls" in deferred.fit.keywords
        assert deferred.fit.keywords["cls"] == sk_preprocessing.StandardScaler

    def test_predictor_uses_fit_sklearn(self):
        """Test LinearRegression uses fit_sklearn with target support."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0], "y": [0.0, 1.0]})
        step = xo.Step.from_instance_name(sk_linear_model.LinearRegression(), name="lr")
        fitted = step.fit(t, features=("a", "b"), target="y")

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # fit_sklearn uses cls and params keywords
        assert "cls" in deferred.fit.keywords
        assert deferred.fit.keywords["cls"] == sk_linear_model.LinearRegression


class TestFromFittedStepOtherFunctions:
    """Tests verifying correct other (transform/predict) function is used."""

    def test_series_kv_encoded_uses_transform_sklearn_series_kv(self):
        """Test TfidfVectorizer uses transform_sklearn_series_kv."""
        t = xo.memtable({"text": ["hello world", "foo bar"]})
        step = xo.Step.from_instance_name(
            sk_feature_extraction_text.TfidfVectorizer(), name="tfidf"
        )
        fitted = step.fit(t, features=("text",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # transform_sklearn_series_kv has col keyword
        assert deferred.other.keywords.get("col") == "text"

    def test_pure_kv_encoded_uses_kv_encode_output(self):
        """Test OneHotEncoder uses kv_encode_output."""
        t = xo.memtable({"cat": ["a", "b", "c"]})
        step = xo.Step.from_instance_name(sk_preprocessing.OneHotEncoder(), name="ohe")
        fitted = step.fit(t, features=("cat",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # kv_encode_output is the function itself (not curried with keywords)
        assert deferred.other is kv_encode_output

    def test_pure_struct_uses_transform_sklearn_struct(self):
        """Test StandardScaler uses transform_sklearn_struct with convert_array."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0]})
        step = xo.Step.from_instance_name(
            sk_preprocessing.StandardScaler(), name="scaler"
        )
        fitted = step.fit(t, features=("a", "b"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # transform_sklearn_struct is the base function (curried with convert_array as first arg)
        # Check that it's transform_sklearn_struct by verifying function name
        assert deferred.other.func.__name__ == "transform_sklearn_struct"


class TestFromFittedStepReturnTypes:
    """Tests verifying correct return types for each transformer type."""

    def test_kv_encoded_return_type(self):
        """Test KV-encoded transformers return KV_ENCODED_TYPE."""
        t = xo.memtable({"cat": ["a", "b", "c"]})
        step = xo.Step.from_instance_name(sk_preprocessing.OneHotEncoder(), name="ohe")
        fitted = step.fit(t, features=("cat",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred.return_type == KV_ENCODED_TYPE
        # KV_ENCODED_TYPE is Array[Struct{key: string, value: float64}]
        assert isinstance(deferred.return_type, dt.Array)

    def test_struct_return_type_has_correct_fields(self):
        """Test struct transformers return type with correct field names."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0]})
        step = xo.Step.from_instance_name(
            sk_preprocessing.StandardScaler(), name="scaler"
        )
        fitted = step.fit(t, features=("a", "b"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert isinstance(deferred.return_type, dt.Struct)
        # Struct should have same field names as features
        field_names = set(deferred.return_type.names)
        assert field_names == {"a", "b"}

    def test_predict_return_type_matches_target(self):
        """Test predict return type matches the target column type."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0], "y": [0.0, 1.0]})
        step = xo.Step.from_instance_name(sk_linear_model.LinearRegression(), name="lr")
        fitted = step.fit(t, features=("a", "b"), target="y")

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # Return type should be float64 to match target
        assert deferred.return_type == dt.float64


class TestFromFittedStepTargetHandling:
    """Tests verifying target is correctly passed through for supervised cases."""

    def test_transformer_with_target_passes_target(self):
        """Test supervised transformer (SelectKBest) preserves target."""
        t = xo.memtable(
            {"a": [1.0, 2.0, 3.0, 4.0], "b": [4.0, 3.0, 2.0, 1.0], "y": [0, 0, 1, 1]}
        )
        step = xo.Step.from_instance_name(
            sk_feature_selection.SelectKBest(k=1), name="skb"
        )
        fitted = step.fit(t, features=("a", "b"), target="y")

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred.target == "y"

    def test_unsupervised_transformer_has_no_target(self):
        """Test unsupervised transformer (StandardScaler) has None target."""
        t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0]})
        step = xo.Step.from_instance_name(
            sk_preprocessing.StandardScaler(), name="scaler"
        )
        fitted = step.fit(t, features=("a", "b"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        assert deferred.target is None

    def test_kv_encoded_transformer_target_cleared_when_not_needed(self):
        """Test KV-encoded transformer clears target when structer doesn't need it."""
        t = xo.memtable({"cat": ["a", "b", "c"]})
        step = xo.Step.from_instance_name(sk_preprocessing.OneHotEncoder(), name="ohe")
        # OneHotEncoder doesn't need target even if provided
        fitted = step.fit(t, features=("cat",))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # Target should be None for unsupervised transformers
        assert deferred.target is None


class TestFromFittedStepColumnTransformerWithKvFields:
    """Tests for ColumnTransformer producing struct with KV-encoded fields."""

    def test_column_transformer_mixed_output_struct_has_kv_fields(self):
        """Test ColumnTransformer with scaler + encoder produces struct with KV fields."""
        t = xo.memtable({"num": [1.0, 2.0, 3.0], "cat": ["a", "b", "c"]})
        ct = sk_compose.ColumnTransformer(
            [
                ("scaler", sk_preprocessing.StandardScaler(), ["num"]),
                ("encoder", sk_preprocessing.OneHotEncoder(), ["cat"]),
            ]
        )
        step = xo.Step.from_instance_name(ct, name="ct")
        fitted = step.fit(t, features=("num", "cat"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # Return type should be struct
        assert isinstance(deferred.return_type, dt.Struct)
        # Struct should have both regular fields and KV-encoded field
        field_types = dict(deferred.return_type.items())
        # scaler output should have regular float fields
        assert "scaler__num" in field_types
        # encoder output should be KV-encoded
        assert "encoder" in field_types
        assert field_types["encoder"] == KV_ENCODED_TYPE

    def test_column_transformer_all_kv_encoded_uses_struct_kv_branch(self):
        """Test ColumnTransformer with all KV-encoded outputs uses struct_kv branch."""
        t = xo.memtable({"cat1": ["a", "b", "c"], "cat2": ["x", "y", "z"]})
        ct = sk_compose.ColumnTransformer(
            [
                ("enc1", sk_preprocessing.OneHotEncoder(), ["cat1"]),
                ("enc2", sk_preprocessing.OneHotEncoder(), ["cat2"]),
            ]
        )
        step = xo.Step.from_instance_name(ct, name="ct")
        fitted = step.fit(t, features=("cat1", "cat2"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # Should use struct_kv branch since struct exists with KV fields
        assert deferred.name_infix == "transformed_struct_kv"
        # All fields should be KV-encoded
        field_types = dict(deferred.return_type.items())
        for name, typ in field_types.items():
            assert typ == KV_ENCODED_TYPE, f"Field {name} should be KV-encoded"

    def test_column_transformer_mixed_executes_correctly(self):
        """Test ColumnTransformer with mixed output produces correct result."""
        t = xo.memtable({"num": [1.0, 2.0, 3.0], "cat": ["a", "b", "a"]})
        ct = sk_compose.ColumnTransformer(
            [
                ("scaler", sk_preprocessing.StandardScaler(), ["num"]),
                ("encoder", sk_preprocessing.OneHotEncoder(), ["cat"]),
            ]
        )
        step = xo.Step.from_instance_name(ct, name="ct")
        fitted = step.fit(t, features=("num", "cat"))

        deferred = DeferredFitOther.from_fitted_step(fitted)

        # Execute the transform
        col = deferred.deferred_other.on_expr(t)
        result = t.mutate(transformed=col).execute()

        # Verify we get struct output
        assert "transformed" in result.columns
        first_row = result["transformed"].iloc[0]
        # Should have scaler__num (float) and encoder (list of dicts)
        assert "scaler__num" in first_row
        assert "encoder" in first_row
        # encoder should be KV-encoded (list of dicts)
        assert isinstance(first_row["encoder"], list)
        assert all("key" in d and "value" in d for d in first_row["encoder"])



from __future__ import annotations

from decimal import Context, Decimal, InvalidOperation


def normalize_decimal(
    value,
    precision: int | None = None,
    scale: int | None = None,
    strict: bool = True,
):
    context = Context(prec=38 if precision is None else precision)

    try:
        if isinstance(value, float):
            out = Decimal(str(value))
        else:
            out = Decimal(value)
    except InvalidOperation:
        raise TypeError(f"Unable to construct decimal from {value!r}")

    out = out.normalize(context=context)
    components = out.as_tuple()
    n_digits = len(components.digits)
    exponent = components.exponent

    if precision is not None and precision < n_digits:
        raise TypeError(
            f"Decimal value {value} has too many digits for precision: {precision}"
        )

    if scale is not None:
        if strict and exponent < -scale:
            raise TypeError(
                f"Normalizing {value} with scale {exponent} to scale -{scale} "
                "would loose precision"
            )

        other = Decimal(10) ** -scale
        try:
            out = out.quantize(other, context=context)
        except InvalidOperation:
            raise TypeError(
                f"Unable to normalize {value!r} as decimal with precision {precision} "
                f"and scale {scale}"
            )

    return out



from pathlib import Path

import dask
import numpy as np
import pandas as pd
import pytest

import xorq.api as xo


# ensure registration of numpy and pandas objects for tokenization purposes
dask.base.normalize_token.dispatch(np.dtype)
dask.base.normalize_token.dispatch(pd.DataFrame)

array_types_df = pd.DataFrame(
    [
        (
            [np.int64(1), 2, 3],
            ["a", "b", "c"],
            [1.0, 2.0, 3.0],
            "a",
            1.0,
            [[], [np.int64(1), 2, 3], None],
        ),
        (
            [4, 5],
            ["d", "e"],
            [4.0, 5.0],
            "a",
            2.0,
            [],
        ),
        (
            [6, None],
            ["f", None],
            [6.0, np.nan],
            "a",
            3.0,
            [None, [], None],
        ),
        (
            [None, 1, None],
            [None, "a", None],
            [],
            "b",
            4.0,
            [[1], [2], [], [3, 4, 5]],
        ),
        (
            [2, None, 3],
            ["b", None, "c"],
            np.nan,
            "b",
            5.0,
            None,
        ),
        (
            [4, None, None, 5],
            ["d", None, None, "e"],
            [4.0, np.nan, np.nan, 5.0],
            "c",
            6.0,
            [[1, 2, 3]],
        ),
    ],
    columns=[
        "x",
        "y",
        "z",
        "grouper",
        "scalar_column",
        "multi_dim",
    ],
)

win = pd.DataFrame(
    {
        "g": ["a", "a", "a", "a", "a"],
        "x": [0, 1, 2, 3, 4],
        "y": [3, 2, 0, 1, 1],
    }
)

expected_tables = (
    "array_types",
    "astronauts",
    "awards_players",
    "awards_players_special_types",
    "batting",
    "diamonds",
    "functional_alltypes",
    "geo",
    "geography_columns",
    "geometry_columns",
    "json_t",
    "map",
    "spatial_ref_sys",
    "topk",
    "tzone",
)


def remove_unexpected_tables(dirty):
    # drop tables
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_table(table, force=True)

    # drop view
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_view(table, force=True)

    actual = sorted(dirty.list_tables())
    expected = sorted(expected_tables)
    if actual != expected:
        missing = tuple(t for t in expected if t not in actual)
        extra = tuple(t for t in actual if t not in expected)
        raise ValueError(
            {
                "missing": missing,
                "extra": extra,
            }
        )


@pytest.fixture(scope="function")
def pg():
    conn = xo.postgres.connect_env()
    remove_unexpected_tables(conn)
    yield conn
    remove_unexpected_tables(conn)


@pytest.fixture(scope="session")
def root_dir():
    return Path(__file__).absolute().parents[2]


@pytest.fixture(scope="session")
def parquet_dir(root_dir):
    data_dir = root_dir / "ci" / "ibis-testing-data" / "parquet"
    return data_dir


@pytest.fixture(scope="session")
def fixture_dir(root_dir):
    return root_dir.joinpath("python", "xorq", "tests", "fixtures")


@pytest.fixture(scope="session")
def data_dir(root_dir):
    data_dir = root_dir / "ci" / "ibis-testing-data"
    return data_dir


@pytest.fixture(scope="session")
def csv_dir(data_dir):
    csv_dir = data_dir / "csv"
    return csv_dir


@pytest.fixture(scope="session")
def examples_dir(root_dir):
    examples_dir = root_dir / "examples"
    return examples_dir



from __future__ import annotations


__all__ = [
    "BigQueryCompiler",
    "ClickHouseCompiler",
    "DataFusionCompiler",
    "DruidCompiler",
    "DuckDBCompiler",
    "ExasolCompiler",
    "FlinkCompiler",
    "ImpalaCompiler",
    "MSSQLCompiler",
    "MySQLCompiler",
    "OracleCompiler",
    "PostgresCompiler",
    "PySparkCompiler",
    "RisingWaveCompiler",
    "SnowflakeCompiler",
    "SQLiteCompiler",
    "TrinoCompiler",
]

from xorq.vendor.ibis.backends.sql.compilers.bigquery import BigQueryCompiler
from xorq.vendor.ibis.backends.sql.compilers.clickhouse import ClickHouseCompiler
from xorq.vendor.ibis.backends.sql.compilers.datafusion import DataFusionCompiler
from xorq.vendor.ibis.backends.sql.compilers.druid import DruidCompiler
from xorq.vendor.ibis.backends.sql.compilers.duckdb import DuckDBCompiler
from xorq.vendor.ibis.backends.sql.compilers.exasol import ExasolCompiler
from xorq.vendor.ibis.backends.sql.compilers.flink import FlinkCompiler
from xorq.vendor.ibis.backends.sql.compilers.impala import ImpalaCompiler
from xorq.vendor.ibis.backends.sql.compilers.mssql import MSSQLCompiler
from xorq.vendor.ibis.backends.sql.compilers.mysql import MySQLCompiler
from xorq.vendor.ibis.backends.sql.compilers.oracle import OracleCompiler
from xorq.vendor.ibis.backends.sql.compilers.postgres import PostgresCompiler
from xorq.vendor.ibis.backends.sql.compilers.pyspark import PySparkCompiler
from xorq.vendor.ibis.backends.sql.compilers.risingwave import RisingWaveCompiler
from xorq.vendor.ibis.backends.sql.compilers.snowflake import SnowflakeCompiler
from xorq.vendor.ibis.backends.sql.compilers.sqlite import SQLiteCompiler
from xorq.vendor.ibis.backends.sql.compilers.trino import TrinoCompiler



import functools
from typing import Any

import cloudpickle
import pyarrow as pa
import toolz

import xorq.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor.ibis.common.annotations import Argument
from xorq.vendor.ibis.common.collections import FrozenDict
from xorq.vendor.ibis.expr.operations import Namespace
from xorq.vendor.ibis.expr.operations.udf import (
    AggUDF,
    InputType,
    ScalarUDF,
    _make_udf_name,
    _wrap,
    reduce_udf,
    scalar,
)
from xorq.vendor.ibis.expr.operations.udf import (
    agg as _agg,
)


def property_wrap_fn(fn):
    return property(fget=lambda _, fn=fn: fn)


def arrays_to_df(names, *arrays):
    import pandas as pd

    return pd.DataFrame(
        {name: array.to_pandas() for (name, array) in zip(names, arrays)}
    )


def make_pyarrow_array(return_type, series):
    return pa.Array.from_pandas(series, type=return_type.to_pyarrow())


def make_dunder_func(fn, schema, return_type=None):
    def fn_from_arrays(*arrays):
        df = arrays_to_df(schema, *arrays)
        value = fn(df)
        if return_type is not None:
            value = make_pyarrow_array(return_type, value)
        return value

    return property_wrap_fn(fn_from_arrays)


def make_expr_scalar_udf_dunder_func(fn, schema, return_type):
    def fn_from_arrays(*arrays, computed_arg=None, **kwargs):
        if computed_arg is None:
            raise ValueError(
                "Caller must bind computed_arg to the output of computed_kwargs_expr"
            )
        df = arrays_to_df(schema, *arrays)
        value = fn(computed_arg, df, **kwargs)
        return make_pyarrow_array(
            return_type,
            value,
        )

    return property_wrap_fn(fn_from_arrays)


@toolz.curry
def wrap_model(value, model_key="model"):
    return cloudpickle.dumps({model_key: value})


unwrap_model = cloudpickle.loads


class ExprScalarUDF(ScalarUDF):
    @property
    def computed_kwargs_expr(self):
        # must push the expr into __config__ so that it doesn't get turned into a window function
        return self.__config__["computed_kwargs_expr"]

    @property
    def post_process_fn(self):
        return self.__config__["post_process_fn"]

    @property
    def schema(self):
        return self.__config__["schema"]

    def on_expr(self, e, **kwargs):
        # rebind deferred_model (computed_kwargs_expr) to a new expr
        return type(self)(*(e[c] for c in self.schema), **kwargs)

    def with_computed_kwargs_expr(self, computed_kwargs_expr):
        # we must create a new typ to set __config__
        restore_udf, (name, bases, type_kwargs, ctor_kwargs) = self.__reduce__()
        new_config = FrozenDict(
            type_kwargs["__config__"] | {"computed_kwargs_expr": computed_kwargs_expr}
        )
        new_type_kwargs = type_kwargs | {"__config__": new_config}
        new_op = restore_udf(name, bases, new_type_kwargs, ctor_kwargs)
        return new_op

    __reduce__ = reduce_udf


@toolz.curry
def make_pandas_expr_udf(
    computed_kwargs_expr,
    fn,
    schema,
    return_type=dt.binary,
    database=None,
    catalog=None,
    name=None,
    *,
    post_process_fn=unwrap_model,
    **kwargs,
):
    """
    Create an expression-based scalar UDF that incorporates pre-computed values.

    This function creates a special type of scalar UDF that can access pre-computed
    values (like trained machine learning models) during execution. The pre-computed
    value is generated from a separate expression and passed to the UDF function,
    enabling complex workflows like model training and inference within the same
    query pipeline.

    Parameters
    ----------
    computed_kwargs_expr : Expression
        An expression that computes a value to be passed to the UDF function.
        This is typically an aggregation that produces a model or other computed value.
    fn : callable
        The function to be executed. Should accept (computed_arg, df, **kwargs)
        where computed_arg is the result of computed_kwargs_expr and df is a
        pandas DataFrame containing the input columns.
    schema : Schema
        The input schema defining column names and their data types.
    return_type : DataType, default dt.binary
        The return data type of the UDF.
    database : str, optional
        Database name for the UDF namespace.
    catalog : str, optional
        Catalog name for the UDF namespace.
    name : str, optional
        Name of the UDF. If None, uses the function name.
    post_process_fn : callable, default unwrap_model
        Function to post-process the computed_kwargs_expr result before passing
        to the main function.
    **kwargs
        Additional configuration parameters.

    Returns
    -------
    callable
        A UDF constructor that can be used in expressions.

    Examples
    --------
    Machine learning workflow with penguin species classification:

    >>> import pickle
    >>> import pandas as pd
    >>> from sklearn.neighbors import KNeighborsClassifier
    >>> from xorq.expr.udf import make_pandas_expr_udf, agg
    >>> import xorq.expr.datatypes as dt
    >>> import xorq.api as xo
    >>>
    >>> # Load penguins dataset
    >>> penguins = xo.examples.penguins.fetch(backend=xo.connect())
    >>> features = ['bill_length_mm', 'bill_depth_mm', 'flipper_length_mm', 'body_mass_g']
    >>>
    >>> # Split data
    >>> train_data = penguins.filter(penguins.year < 2009)
    >>> test_data = penguins.filter(penguins.year >= 2009)
    >>>
    >>> # Define training function
    >>> def train_penguin_model(df):
    ...     df_clean = df.dropna(subset=features + ['species'])
    ...     X = df_clean[features]
    ...     y = df_clean['species']
    ...
    ...     model = KNeighborsClassifier(n_neighbors=5)
    ...     model.fit(X, y)
    ...     return pickle.dumps(model)
    >>>
    >>> # Define prediction function
    >>> def predict_penguin_species(model, df):
    ...     df_clean = df.dropna(subset=features)
    ...     X = df_clean[features]
    ...     predictions = model.predict(X)
    ...     # Return predictions for all rows (fill NaN for missing data)
    ...     result = pd.Series(index=df.index, dtype='object')
    ...     result.loc[df_clean.index] = predictions
    ...     return result.fillna('Unknown')
    >>>
    >>> # Create schemas for training and prediction
    >>> train_schema = train_data.select(features + ['species']).schema()
    >>> test_schema = test_data.select(features).schema()
    >>>
    >>> # Create model training UDAF
    >>> model_udaf = agg.pandas_df(
    ...     fn=train_penguin_model,
    ...     schema=train_schema,
    ...     return_type=dt.binary,
    ...     name="train_penguin_model"
    >>> )
    >>>
    >>> # Create prediction UDF that uses trained model
    >>> predict_udf = make_pandas_expr_udf(
    ...     computed_kwargs_expr=model_udaf.on_expr(train_data),
    ...     fn=predict_penguin_species,
    ...     schema=test_schema,
    ...     return_type=dt.string,
    ...     name="predict_species"
    >>> )
    >>>
    >>> # Apply predictions to test data
    >>> result = test_data.mutate(
    ...     predicted_species=predict_udf.on_expr(test_data)
    >>> ).execute()

    Penguin size classification with pre-computed thresholds:

    >>> def compute_size_thresholds(df):
    ...     df_clean = df.dropna(subset=['body_mass_g'])
    ...     return {
    ...         'small_threshold': df_clean['body_mass_g'].quantile(0.33),
    ...         'large_threshold': df_clean['body_mass_g'].quantile(0.67)
    ...     }
    >>>
    >>> def classify_penguin_size(thresholds, df):
    ...     def classify_size(mass):
    ...         if pd.isna(mass):
    ...             return 'Unknown'
    ...         elif mass < thresholds['small_threshold']:
    ...             return 'Small'
    ...         elif mass > thresholds['large_threshold']:
    ...             return 'Large'
    ...         else:
    ...             return 'Medium'
    ...
    ...     return df['body_mass_g'].apply(classify_size)
    >>>
    >>> # Create threshold computation UDAF
    >>> threshold_udaf = agg.pandas_df(
    ...     fn=compute_size_thresholds,
    ...     schema=penguins.select(['body_mass_g']).schema(),
    ...     return_type=dt.Struct({
    ...         'small_threshold': dt.float64,
    ...         'large_threshold': dt.float64
    ...     }),
    ...     name="compute_thresholds"
    >>> )
    >>>
    >>> # Create size classification UDF
    >>> size_classify_udf = make_pandas_expr_udf(
    ...     computed_kwargs_expr=threshold_udaf.on_expr(penguins),
    ...     fn=classify_penguin_size,
    ...     schema=penguins.select(['body_mass_g']).schema(),
    ...     return_type=dt.string,
    ...     name="classify_size",
    ...     post_process_fn=lambda x: x  # thresholds are already a dict
    >>> )
    >>>
    >>> # Apply size classification
    >>> result = penguins.mutate(
    ...     size_category=size_classify_udf.on_expr(penguins)
    >>> ).execute()

    Notes
    -----
    This UDF type is particularly powerful for ML workflows where you need to:
    1. Train a model on aggregated data
    2. Serialize the trained model
    3. Use the model for predictions on new data

    The computed_kwargs_expr is evaluated once and its result is passed to every
    invocation of the main function, enabling efficient model reuse.

    See Also
    --------
    make_pandas_udf : For standard pandas-based scalar UDFs
    agg : For aggregation functions
    """

    name = name if name is not None else _make_udf_name(fn)
    bases = (ExprScalarUDF,)
    fields = {
        arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
        for (arg_name, typ) in schema.items()
    }
    meta = {
        "dtype": return_type,
        "__input_type__": InputType.PYARROW,
        "__func__": make_expr_scalar_udf_dunder_func(fn, schema, return_type),
        # valid config keys: computed_kwargs_expr, post_process_fn, volatility
        "__config__": FrozenDict(
            computed_kwargs_expr=computed_kwargs_expr,
            post_process_fn=post_process_fn,
            schema=schema,
            fn=fn,
            **kwargs,
        ),
        "__udf_namespace__": Namespace(database=database, catalog=catalog),
        "__module__": fn.__module__,
        # FIXME: determine why this fails with case mismatch by default
        "__func_name__": name,
    }
    kwds = {
        **fields,
        **meta,
    }

    node = type(
        name,
        bases,
        kwds,
    )

    # FIXME: enable working with deferred like _wrap enables
    @functools.wraps(fn)
    def construct(*args: Any, **kwargs: Any) -> ir.Value:
        return node(*args, **kwargs).to_expr()

    def on_expr(e, **kwargs):
        return construct(*(e[c] for c in schema), **kwargs)

    construct.on_expr = on_expr
    construct.fn = fn
    construct.return_type = return_type
    return construct


@toolz.curry
def make_pandas_udf(
    fn, schema, return_type, database=None, catalog=None, name=None, **kwargs
):
    """
    Create a scalar User-Defined Function (UDF) that operates on pandas DataFrames.

    This function creates a scalar UDF that processes data row-by-row, converting
    PyArrow arrays to pandas DataFrames for processing. It's ideal for operations
    that benefit from pandas' rich functionality and are easier to express with
    DataFrame operations.

    Parameters
    ----------
    fn : callable
        The function to be executed. Should accept a pandas DataFrame and return
        a pandas Series or scalar value.
    schema : Schema
        The input schema defining column names and their data types.
    return_type : DataType
        The return data type of the UDF.
    database : str, optional
        Database name for the UDF namespace.
    catalog : str, optional
        Catalog name for the UDF namespace.
    name : str, optional
        Name of the UDF. If None, generates a name from the function.
    **kwargs
        Additional configuration parameters (e.g., volatility settings).

    Returns
    -------
    callable
        A UDF constructor that can be used in expressions with `.on_expr()` method.

    Examples
    --------
    Creating a UDF that calculates penguin bill ratio:

    >>> import pandas as pd
    >>> from xorq.expr.udf import make_pandas_udf
    >>> import xorq.expr.datatypes as dt
    >>> import xorq.api as xo
    >>>
    >>> # Load penguins dataset
    >>> penguins = xo.examples.penguins.fetch(backend=xo.connect())
    >>>
    >>> # Define the function
    >>> def bill_ratio(df):
    ...     return df['bill_length_mm'] / df['bill_depth_mm']
    >>>
    >>> # Create UDF
    >>> schema = penguins.select(['bill_length_mm', 'bill_depth_mm']).schema()
    >>> bill_ratio_udf = make_pandas_udf(
    ...     fn=bill_ratio,
    ...     schema=schema,
    ...     return_type=dt.float64,
    ...     name="bill_ratio"
    >>> )
    >>>
    >>> # Apply to table
    >>> result = penguins.mutate(
    ...     bill_ratio=bill_ratio_udf.on_expr(penguins)
    >>> ).execute()

    Creating a UDF for penguin size classification:

    >>> def classify_penguin_size(df):
    ...     def size_category(row):
    ...         mass = row['body_mass_g']
    ...         flipper = row['flipper_length_mm']
    ...
    ...         if pd.isna(mass) or pd.isna(flipper):
    ...             return 'Unknown'
    ...
    ...         # Simple size classification based on body mass and flipper length
    ...         if mass > 4500 and flipper > 210:
    ...             return 'Large'
    ...         elif mass < 3500 and flipper < 190:
    ...             return 'Small'
    ...         else:
    ...             return 'Medium'
    ...
    ...     return df.apply(size_category, axis=1)
    >>>
    >>> size_schema = penguins.select(['body_mass_g', 'flipper_length_mm']).schema()
    >>> size_udf = make_pandas_udf(
    ...     fn=classify_penguin_size,
    ...     schema=size_schema,
    ...     return_type=dt.string,
    ...     name="classify_size"
    >>> )
    >>>
    >>> # Apply size classification
    >>> result = penguins.mutate(
    ...     size_category=size_udf.on_expr(penguins)
    >>> ).execute()

    Creating a UDF for complex penguin feature engineering:

    >>> def penguin_features(df):
    ...     # Create multiple derived features
    ...     features = pd.DataFrame(index=df.index)
    ...
    ...     # Bill area
    ...     features['bill_area'] = df['bill_length_mm'] * df['bill_depth_mm']
    ...
    ...     # Body condition index
    ...     features['body_condition'] = df['body_mass_g'] / (df['flipper_length_mm'] ** 2)
    ...
    ...     # Aspect ratio of bill
    ...     features['bill_aspect_ratio'] = df['bill_length_mm'] / df['bill_depth_mm']
    ...
    ...     # Return as concatenated string for this example
    ...     return features.apply(lambda row: f"area:{row['bill_area']:.1f}_bci:{row['body_condition']:.4f}_ratio:{row['bill_aspect_ratio']:.2f}", axis=1)
    >>>
    >>> all_measurements = ['bill_length_mm', 'bill_depth_mm', 'flipper_length_mm', 'body_mass_g']
    >>> features_schema = penguins.select(all_measurements).schema()
    >>> features_udf = make_pandas_udf(
    ...     fn=penguin_features,
    ...     schema=features_schema,
    ...     return_type=dt.string,
    ...     name="penguin_features"
    >>> )
    >>>
    >>> # Apply feature engineering
    >>> result = penguins.mutate(
    ...     derived_features=features_udf.on_expr(penguins)
    >>> ).execute()

    Notes
    -----
    - The function receives a pandas DataFrame where columns correspond to the schema keys
    - The function should return a pandas Series or scalar value compatible with return_type
    - PyArrow arrays are automatically converted to pandas and back for seamless integration
    - Use this when you need pandas-specific functionality like string operations,
      datetime handling, or complex data manipulations

    See Also
    --------
    scalar : For PyArrow-based scalar UDFs with potentially better performance
    make_pandas_expr_udf : For UDFs that need pre-computed values
    agg : For aggregation functions
    """

    from xorq.vendor.ibis.expr.operations.udf import ScalarUDF

    name = name if name is not None else _make_udf_name(fn)
    bases = (ScalarUDF,)
    fields = {
        arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
        for (arg_name, typ) in schema.items()
    }
    meta = {
        "dtype": return_type,
        "__input_type__": InputType.PYARROW,
        "__func__": make_dunder_func(fn, schema, return_type),
        # valid config keys: volatility
        "__config__": FrozenDict(**kwargs),
        "__udf_namespace__": Namespace(database=database, catalog=catalog),
        "__module__": fn.__module__,
        # FIXME: determine why this fails with case mismatch by default
        "__func_name__": name,
    }
    kwds = {
        **fields,
        **meta,
    }

    node = type(
        name,
        bases,
        kwds,
    )

    # FIXME: enable working with deferred like _wrap enables
    @functools.wraps(fn)
    def construct(*args: Any, **kwargs: Any) -> ir.Value:
        return node(*args, **kwargs).to_expr()

    def on_expr(e, **kwargs):
        return construct(*(e[c] for c in schema), **kwargs)

    construct.on_expr = on_expr
    construct.fn = fn
    return construct


class agg(_agg):
    __slots__ = ()

    _base = AggUDF

    @classmethod
    def pyarrow(
        cls,
        fn=None,
        name=None,
        signature=None,
        **kwargs,
    ):
        """
        Decorator for creating PyArrow-based aggregation functions.

        This method creates high-performance aggregation UDFs that operate directly
        on PyArrow arrays using PyArrow compute functions. It's ideal for numerical
        computations and operations that benefit from vectorized processing.

        Parameters
        ----------
        fn : callable, optional
            The aggregation function. Should accept PyArrow arrays and return a scalar
            or array result using PyArrow compute functions.
        name : str, optional
            Name of the UDF. If None, uses the function name.
        signature : Signature, optional
            Function signature specification for type checking.
        **kwargs
            Additional configuration parameters like volatility settings.

        Returns
        -------
        callable
            A UDF decorator that can be applied to functions, or if fn is provided,
            the wrapped UDF function.

        Examples
        --------
        Creating a PyArrow aggregation for penguin bill measurements:

        >>> import pyarrow.compute as pc
        >>> from xorq.expr.udf import agg
        >>> import xorq.expr.datatypes as dt
        >>> import xorq.api as xo
        >>>
        >>> # Load penguins dataset
        >>> penguins = xo.examples.penguins.fetch(backend=xo.connect())
        >>>
        >>> @agg.pyarrow
        >>> def bill_length_range(arr: dt.float64) -> dt.float64:
        ...     return pc.subtract(pc.max(arr), pc.min(arr))
        >>>
        >>> # Calculate bill length range by species
        >>> result = penguins.group_by("species").agg(
        ...     length_range=bill_length_range(penguins.bill_length_mm)
        >>> ).execute()

        Creating a weighted average for penguin measurements:

        >>> @agg.pyarrow
        >>> def weighted_avg_flipper(lengths: dt.float64, weights: dt.float64) -> dt.float64:
        ...     return pc.divide(
        ...         pc.sum(pc.multiply(lengths, weights)),
        ...         pc.sum(weights)
        ...     )
        >>>
        >>> # Weighted average flipper length by body mass
        >>> result = penguins.group_by("species").agg(
        ...     weighted_flipper=weighted_avg_flipper(
        ...         penguins.flipper_length_mm,
        ...         penguins.body_mass_g
        ...     )
        >>> ).execute()

        Notes
        -----
        - PyArrow UDAFs typically offer the best performance for numerical operations
        - Functions receive PyArrow arrays and should use PyArrow compute functions
        - The return type should be compatible with PyArrow scalar types
        - Use this for high-performance aggregations on large datasets

        See Also
        --------
        agg.pandas_df : For pandas DataFrame-based aggregations
        agg.builtin : For database-native aggregate functions
        """
        result = _wrap(
            cls._make_wrapper,
            InputType.PYARROW,
            fn,
            name=name,
            signature=signature,
            **kwargs,
        )
        return result

    @classmethod
    @toolz.curry
    def pandas_df(
        cls,
        fn,
        schema,
        return_type,
        database=None,
        catalog=None,
        name=None,
        **kwargs,
    ):
        """
        Create a pandas DataFrame-based aggregation function.

        This method creates aggregation UDFs that operate on pandas DataFrames,
        providing access to the full pandas ecosystem for complex aggregations.
        It's particularly useful for statistical operations, machine learning
        model training, and complex data transformations that are easier to
        express with pandas.

        Parameters
        ----------
        fn : callable
            The aggregation function. Should accept a pandas DataFrame and return
            a value compatible with the return_type. The DataFrame contains all
            columns specified in the schema for each group.
        schema : Schema or dict
            Input schema defining column names and their data types.
        return_type : DataType
            The return data type of the aggregation.
        database : str, optional
            Database name for the UDF namespace.
        catalog : str, optional
            Catalog name for the UDF namespace.
        name : str, optional
            Name of the UDF. If None, generates a name from the function.
        **kwargs
            Additional configuration parameters (e.g., volatility settings).

        Returns
        -------
        callable
            A UDF constructor that can be used in aggregation expressions.

        Examples
        --------
        Training a KNN classifier on penguin data as an aggregation:

        >>> import pickle
        >>> from sklearn.neighbors import KNeighborsClassifier
        >>> from xorq.expr.udf import agg
        >>> import xorq.expr.datatypes as dt
        >>> import xorq.api as xo
        >>>
        >>> # Load penguins dataset
        >>> penguins = xo.examples.penguins.fetch(backend=xo.connect())
        >>> features = ['bill_length_mm', 'bill_depth_mm', 'flipper_length_mm', 'body_mass_g']
        >>>
        >>> def train_penguin_classifier(df):
        ...     # Remove rows with missing values
        ...     df_clean = df.dropna(subset=features + ['species'])
        ...     X = df_clean[features]
        ...     y = df_clean['species']
        ...
        ...     model = KNeighborsClassifier(n_neighbors=3)
        ...     model.fit(X, y)
        ...     return pickle.dumps(model)
        >>>
        >>> # Create the aggregation UDF
        >>> penguin_schema = penguins.select(features + ['species']).schema()
        >>> train_model_udf = agg.pandas_df(
        ...     fn=train_penguin_classifier,
        ...     schema=penguin_schema,
        ...     return_type=dt.binary,
        ...     name="train_penguin_classifier"
        >>> )
        >>>
        >>> # Train one model per island
        >>> trained_models = penguins.group_by("island").agg(
        ...     model=train_model_udf.on_expr(penguins)
        >>> ).execute()

        Complex statistical aggregation for penguin measurements:

        >>> def penguin_stats(df):
        ...     return {
        ...         'bill_ratio_mean': (df['bill_length_mm'] / df['bill_depth_mm']).mean(),
        ...         'mass_flipper_corr': df['body_mass_g'].corr(df['flipper_length_mm']),
        ...         'count': len(df),
        ...         'size_score': (df['body_mass_g'] * df['flipper_length_mm']).mean()
        ...     }
        >>>
        >>> stats_schema = penguins.select([
        ...     'bill_length_mm', 'bill_depth_mm', 'body_mass_g', 'flipper_length_mm'
        >>> ]).schema()
        >>>
        >>> stats_udf = agg.pandas_df(
        ...     fn=penguin_stats,
        ...     schema=stats_schema,
        ...     return_type=dt.Struct({
        ...         'bill_ratio_mean': dt.float64,
        ...         'mass_flipper_corr': dt.float64,
        ...         'count': dt.int64,
        ...         'size_score': dt.float64
        ...     }),
        ...     name="penguin_stats"
        >>> )
        >>>
        >>> # Calculate statistics by species
        >>> result = penguins.group_by("species").agg(
        ...     stats=stats_udf.on_expr(penguins)
        >>> ).execute()

        Feature selection for penguin classification:

        >>> def select_best_penguin_features(df, n_features=2):
        ...     from sklearn.feature_selection import mutual_info_classif
        ...     import pandas as pd
        ...
        ...     df_clean = df.dropna()
        ...     features = ['bill_length_mm', 'bill_depth_mm', 'flipper_length_mm', 'body_mass_g']
        ...     X = df_clean[features]
        ...     y = df_clean['species']
        ...
        ...     scores = mutual_info_classif(X, y)
        ...     return list(pd.Series(scores, index=features).nlargest(n_features).index)
        >>>
        >>> feature_selector = agg.pandas_df(
        ...     fn=select_best_penguin_features,
        ...     schema=penguins.schema(),
        ...     return_type=dt.Array(dt.string),
        ...     name="select_penguin_features"
        >>> )
        >>>
        >>> # Find best features by island
        >>> best_features = penguins.group_by("island").agg(
        ...     top_features=feature_selector.on_expr(penguins)
        >>> ).execute()

        Notes
        -----
        - The function receives a pandas DataFrame containing all rows in each group
        - PyArrow arrays are automatically converted to pandas for processing
        - The function can return complex data structures (dicts, lists) if return_type supports it
        - Use this when you need pandas-specific functionality or ML libraries
        - Performance may be lower than PyArrow UDAFs for simple numerical operations
        - Particularly powerful for ML model training workflows

        See Also
        --------
        agg.pyarrow : For high-performance PyArrow-based aggregations
        make_pandas_expr_udf : For using trained models in prediction UDFs
        make_pandas_udf : For scalar pandas operations
        """
        name = name if name is not None else fn.__name__
        bases = (cls._base,)
        fields = {
            arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
            for (arg_name, typ) in schema.items()
        }
        meta = {
            "dtype": return_type,
            "__input_type__": InputType.PYARROW,
            "__func__": make_dunder_func(fn, schema),
            # valid config keys: volatility
            "__config__": FrozenDict(fn=fn, **kwargs),
            "__udf_namespace__": Namespace(database=database, catalog=catalog),
            "__module__": fn.__module__,
            # FIXME: determine why this fails with case mismatch by default
            "__func_name__": name,
        }
        kwds = {
            **fields,
            **meta,
        }
        node = type(
            _make_udf_name(fn.__name__),
            bases,
            kwds,
        )

        # FIXME: enable working with deferred like _wrap enables
        @functools.wraps(fn)
        def construct(*args: Any, **kwargs: Any) -> ir.Value:
            return node(*args, **kwargs).to_expr()

        def on_expr(e, **kwargs):
            return construct(*(e[c] for c in schema), **kwargs)

        construct.on_expr = on_expr
        construct.fn = fn
        return construct


def arbitrate_evaluate(
    uses_window_frame=False,
    supports_bounded_execution=False,
    include_rank=False,
    **config_kwargs,
):
    match (uses_window_frame, supports_bounded_execution, include_rank):
        case (False, False, False):
            return "evaluate_all"
        case (False, True, False):
            return "evaluate"
        case (False, _, True):
            return "evaluate_all_with_rank"
        case (True, _, _):
            return "evaluate"
        case _:
            raise RuntimeError


@toolz.curry
def pyarrow_udwf(
    fn,
    schema,
    return_type,
    name=None,
    namespace=Namespace(database=None, catalog=None),  # noqa: B008
    base=AggUDF,
    **config_kwargs,
):
    """
    Create a User-Defined Window Function (UDWF) using PyArrow.

    This decorator creates window functions that can process partitions of data with
    support for ordering, framing, and ranking. UDWFs are powerful for implementing
    custom analytics functions that need to operate over ordered sets of data within
    partitions.

    Parameters
    ----------
    fn : callable
        The window function implementation. The signature depends on config_kwargs:

        - Basic window function: `fn(self, values: list[pa.Array], num_rows: int) -> pa.Array`
        - With window frame: `fn(self, values: list[pa.Array], eval_range: tuple[int, int]) -> pa.Scalar`
        - With ranking: `fn(self, num_rows: int, ranks_in_partition: list[tuple[int, int]]) -> pa.Array`

    schema : Schema
        Input schema defining column names and data types.
    return_type : DataType
        The return data type of the window function.
    name : str, optional
        Name of the UDWF. If None, uses the function name.
    namespace : Namespace, optional
        Database and catalog namespace for the function.
    base : class, default AggUDF
        Base class for the UDWF (typically AggUDF).
    **config_kwargs
        Configuration options:

        - `uses_window_frame` (bool): Whether function uses window framing
        - `supports_bounded_execution` (bool): Whether function supports bounded execution
        - `include_rank` (bool): Whether function uses ranking information
        - Custom parameters: Additional parameters accessible via `self` in the function

    Returns
    -------
    callable
        A UDWF constructor that can be used in window expressions.

    Examples
    --------
    Exponential smoothing for penguin body mass by species:

    >>> from xorq.expr.udf import pyarrow_udwf
    >>> import pyarrow as pa
    >>> import xorq.api as xo
    >>> import xorq.expr.datatypes as dt
    >>> from xorq.vendor import ibis
    >>>
    >>> # Load penguins dataset
    >>> penguins = xo.examples.penguins.fetch(backend=xo.connect())
    >>>
    >>> @pyarrow_udwf(
    ...     schema=ibis.schema({"body_mass_g": dt.float64}),
    ...     return_type=dt.float64,
    ...     alpha=0.8  # Custom smoothing parameter
    ... )
    >>> def smooth_body_mass(self, values: list[pa.Array], num_rows: int) -> pa.Array:
    ...     results = []
    ...     curr_value = 0.0
    ...     mass_values = values[0]  # body_mass_g column
    ...
    ...     for idx in range(num_rows):
    ...         if idx == 0:
    ...             curr_value = float(mass_values[idx].as_py() or 0)
    ...         else:
    ...             new_val = float(mass_values[idx].as_py() or curr_value)
    ...             curr_value = new_val * self.alpha + curr_value * (1.0 - self.alpha)
    ...         results.append(curr_value)
    ...
    ...     return pa.array(results)
    >>>
    >>> # Apply smoothing within each species, ordered by year
    >>> result = penguins.mutate(
    ...     smooth_mass=smooth_body_mass.on_expr(penguins).over(
    ...         ibis.window(group_by="species", order_by="year")
    ...     )
    >>> ).execute()

    Running difference in penguin bill measurements:

    >>> @pyarrow_udwf(
    ...     schema=ibis.schema({"bill_length_mm": dt.float64}),
    ...     return_type=dt.float64,
    ...     uses_window_frame=True
    ... )
    >>> def bill_length_diff(self, values: list[pa.Array], eval_range: tuple[int, int]) -> pa.Scalar:
    ...     start, stop = eval_range
    ...     bill_values = values[0]
    ...
    ...     if start == stop - 1:  # Single row
    ...         return pa.scalar(0.0)
    ...
    ...     current_val = bill_values[stop - 1].as_py() or 0
    ...     previous_val = bill_values[start].as_py() or 0
    ...     return pa.scalar(float(current_val - previous_val))
    >>>
    >>> # Calculate difference from previous measurement within species
    >>> result = penguins.mutate(
    ...     bill_diff=bill_length_diff.on_expr(penguins).over(
    ...         ibis.window(
    ...             group_by="species",
    ...             order_by="year",
    ...             preceding=1,
    ...             following=0
    ...         )
    ...     )
    >>> ).execute()

    Penguin ranking within species by body mass:

    >>> @pyarrow_udwf(
    ...     schema=ibis.schema({"body_mass_g": dt.float64}),
    ...     return_type=dt.float64,
    ...     include_rank=True
    ... )
    >>> def mass_rank_score(self, num_rows: int, ranks_in_partition: list[tuple[int, int]]) -> pa.Array:
    ...     results = []
    ...     for idx in range(num_rows):
    ...         # Find rank for current row
    ...         rank = next(
    ...             i + 1 for i, (start, end) in enumerate(ranks_in_partition)
    ...             if start <= idx < end
    ...         )
    ...         # Convert rank to score (higher rank = higher score)
    ...         score = 1.0 - (rank - 1) / len(ranks_in_partition)
    ...         results.append(score)
    ...     return pa.array(results)
    >>>
    >>> # Calculate mass rank score within each species
    >>> result = penguins.mutate(
    ...     mass_rank_score=mass_rank_score.on_expr(penguins).over(
    ...         ibis.window(group_by="species", order_by="body_mass_g")
    ...     )
    >>> ).execute()

    Complex penguin feature calculation across measurements:

    >>> @pyarrow_udwf(
    ...     schema=ibis.schema({
    ...         "bill_length_mm": dt.float64,
    ...         "bill_depth_mm": dt.float64,
    ...         "flipper_length_mm": dt.float64
    ...     }),
    ...     return_type=dt.float64,
    ...     window_size=3  # Custom parameter for moving average
    ... )
    >>> def penguin_size_trend(self, values: list[pa.Array], num_rows: int) -> pa.Array:
    ...     bill_length = values[0]
    ...     bill_depth = values[1]
    ...     flipper_length = values[2]
    ...
    ...     results = []
    ...     window_size = self.window_size
    ...
    ...     for idx in range(num_rows):
    ...         # Calculate size metric for current and surrounding rows
    ...         start_idx = max(0, idx - window_size // 2)
    ...         end_idx = min(num_rows, idx + window_size // 2 + 1)
    ...
    ...         size_metrics = []
    ...         for i in range(start_idx, end_idx):
    ...             if (bill_length[i].is_valid and bill_depth[i].is_valid and
    ...                 flipper_length[i].is_valid):
    ...                 # Composite size metric
    ...                 bill_area = bill_length[i].as_py() * bill_depth[i].as_py()
    ...                 size_metric = bill_area * flipper_length[i].as_py()
    ...                 size_metrics.append(size_metric)
    ...
    ...         # Average size metric in window
    ...         avg_size = sum(size_metrics) / len(size_metrics) if size_metrics else 0.0
    ...         results.append(avg_size)
    ...
    ...     return pa.array(results)
    >>>
    >>> # Apply size trend calculation within each species
    >>> result = penguins.mutate(
    ...     size_trend=penguin_size_trend.on_expr(penguins).over(
    ...         ibis.window(group_by="species", order_by="year")
    ...     )
    >>> ).execute()

    Notes
    -----
    The function signature and behavior changes based on configuration:

    - **Standard window function**: Processes all rows in partition, returns array
    - **Frame-based**: Processes specific row ranges, returns scalar per invocation
    - **Rank-aware**: Has access to ranking information within partition

    Custom parameters passed in config_kwargs are accessible as `self.parameter_name`
    in the function implementation.

    See Also
    --------
    scalar : For row-by-row processing
    agg : For aggregation across groups
    make_pandas_udf : For pandas-based scalar operations
    """

    fields = {
        arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
        for (arg_name, typ) in schema.items()
    }
    # which_evaluate = arbitrate_evaluate(**config_kwargs)
    name = name or fn.__name__
    meta = {
        "dtype": return_type,
        "__input_type__": InputType.PYARROW,
        "__func__": property(fget=toolz.functoolz.return_none),
        "__config__": FrozenDict(
            input_types=tuple(datatype for datatype in schema.fields.values()),
            return_type=return_type,
            name=name,
            **config_kwargs,
            # assert which_evaluate in ("evaluate", "evaluate_all", "evaluate_all_with_rank")
            # **{which_evaluate: fn},
            **dict.fromkeys(("evaluate", "evaluate_all", "evaluate_all_with_rank"), fn),
        ),
        "__udf_namespace__": namespace,
        "__module__": __name__,
        "__func_name__": name,
    }
    node = type(
        name,
        (base,),
        {
            **fields,
            **meta,
        },
    )

    def construct(*args: Any, **kwargs: Any) -> ir.Value:
        return node(*args, **kwargs).to_expr()

    def on_expr(e, **kwargs):
        return construct(*(e[c] for c in schema), **kwargs)

    construct.on_expr = on_expr
    return construct


__all__ = [
    "pyarrow_udwf",
    "make_pandas_expr_udf",
    "make_pandas_udf",
    "scalar",
    "agg",
]



from __future__ import annotations

from xorq.vendor.ibis.formats.pandas import PandasData


class PostgresPandasData(PandasData):
    @classmethod
    def convert_GeoSpatial(cls, s, dtype, pandas_type):
        import geopandas as gpd
        import shapely as shp

        return gpd.GeoSeries(shp.from_wkb(s.map(bytes, na_action="ignore")))

    convert_Point = convert_LineString = convert_Polygon = convert_MultiLineString = (
        convert_MultiPoint
    ) = convert_MultiPolygon = convert_GeoSpatial

    @classmethod
    def convert_Binary(cls, s, dtype, pandas_type):
        return s.map(bytes, na_action="ignore")



import functools
import pickle

import toolz
from attr import (
    field,
    frozen,
)
from attr.validators import (
    deep_iterable,
    instance_of,
    optional,
)
from dask.utils import Dispatch
from toolz import compose
from toolz.curried import (
    excepts as cexcepts,
)

import xorq.expr.datatypes as dt
from xorq.backends.xorq import connect
from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
    ParquetTTLSnapshotCache,
    SourceCache,
)
from xorq.common.utils.attr_utils import (
    convert_sorted_kwargs_tuple,
    validate_kwargs_tuple,
)
from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    normalize_attrs,
)
from xorq.common.utils.func_utils import (
    return_constant,
)
from xorq.common.utils.name_utils import (
    make_name,
)
from xorq.expr.ml.enums import ResponseMethod
from xorq.expr.ml.fit_lib import (
    DeferredFitOther,
    decision_function_sklearn,
    feature_importances_sklearn,
    predict_proba_sklearn,
)
from xorq.expr.ml.structer import (
    Structer,
)
from xorq.ibis_yaml.utils import freeze
from xorq.vendor.ibis.expr.types.core import Expr


def do_into_backend(expr, con=None):
    return expr.into_backend(con or connect())


def make_estimator_typ(fit, return_type, name=None, *, transform=None, predict=None):
    from sklearn.base import BaseEstimator

    def arbitrate_transform_predict(transform, predict):
        match (transform, predict):
            case [None, None]:
                raise ValueError(
                    "at least one of transform or predict must be provided"
                )
            case [other, None]:
                return other, "transform"
            case [None, other]:
                return other, ResponseMethod.PREDICT
            case [other0, other1]:
                raise ValueError(other0, other1)
            case _:
                raise ValueError(
                    f"unexpected transform/predict combination: ({transform!r}, {predict!r})"
                )

    assert isinstance(return_type, dt.DataType)
    other, which = arbitrate_transform_predict(transform, predict)
    assert callable(fit) and callable(other)

    def wrapped_fit(self, *args, **kwargs):
        self._model = fit(*args, **kwargs)

    def wrapped_other(self, *args, **kwargs):
        return other(self._model, *args, **kwargs)

    name = name or make_name("estimator", (fit, other))
    typ = type(
        name,
        (BaseEstimator,),
        {
            "fit": wrapped_fit,
            which: wrapped_other,
            "return_type": return_type,
            "_fit": fit,
            f"_{which}": other,
        },
    )
    return typ


@frozen
class Step:
    """
    A single step in a machine learning pipeline that wraps a scikit-learn estimator.

    This class represents an individual processing step that can either transform data
    (transformers like StandardScaler, SelectKBest) or make predictions (classifiers
    like KNeighborsClassifier, LinearSVC). Steps can be combined into Pipeline objects
    to create complex ML workflows.

    Parameters
    ----------
    typ : type
        The scikit-learn estimator class (must inherit from BaseEstimator).
    name : str, optional
        A unique name for this step. If None, generates a name from the class name and ID.
    params_tuple : tuple, optional
        Tuple of (parameter_name, parameter_value) pairs for the estimator.
        Parameters are automatically sorted for consistency.

    Attributes
    ----------
    typ : type
        The scikit-learn estimator class.
    name : str
        The unique name for this step in the pipeline.
    params_tuple : tuple
        Sorted tuple of parameter key-value pairs.

    Examples
    --------
    Create a scaler step:

    >>> from xorq.ml import Step
    >>> from sklearn.preprocessing import StandardScaler
    >>> scaler_step = Step(typ=StandardScaler, name="scaler")
    >>> scaler_step.instance
    StandardScaler()

    Create a classifier step with parameters:

    >>> from sklearn.neighbors import KNeighborsClassifier
    >>> knn_step = Step(
    ...     typ=KNeighborsClassifier,
    ...     name="knn",
    ...     params_tuple=(("n_neighbors", 5), ("weights", "uniform"))
    ... )
    >>> knn_step.instance
    KNeighborsClassifier(n_neighbors=5)

    Notes
    -----
    - The Step class is frozen (immutable) using attrs.
    - All estimators must inherit from sklearn.base.BaseEstimator.
    - Parameter tuples are automatically sorted for hash consistency.
    - Steps can be fitted to data using the fit() method which returns a FittedStep.
    """

    typ = field(validator=instance_of(type))
    name = field(validator=optional(instance_of(str)), default=None)
    params_tuple = field(
        validator=validate_kwargs_tuple,
        converter=compose(freeze, convert_sorted_kwargs_tuple),
        factory=tuple,
    )

    def __attrs_post_init__(self):
        from sklearn.base import BaseEstimator

        assert BaseEstimator in self.typ.mro()
        if self.name is None:
            object.__setattr__(self, "name", f"{self.typ.__name__.lower()}_{id(self)}")

    @property
    def tag_kwargs(self):
        return {name: getattr(self, name) for name in ("typ", "name", "params_tuple")}

    @property
    def instance(self):
        """
        Create an instance of the estimator with the configured parameters.

        Returns
        -------
        object
            An instantiated scikit-learn estimator.
        """

        return self.typ(**dict(self.params_tuple))

    def fit(self, expr, features=None, target=None, cache=None, dest_col=None):
        """
        Fit this step to the given expression data.

        Parameters
        ----------
        expr : Expr
            The xorq expression containing the training data.
        features : tuple of str, optional
            Column names to use as features. If None, infers from expr.columns.
        target : str, optional
            Target column name. Required for prediction steps.
        cache : Cache, optional
            Storage backend for caching fitted models.
        dest_col : str, optional
            Destination column name for transformed output.

        Returns
        -------
        FittedStep
            A fitted step that can transform or predict on new data.
        """

        # how does ColumnTransformer interact with features/target?
        # features = features or self.features or tuple(expr.columns)
        features = features or tuple(expr.columns)
        return FittedStep(
            step=self,
            expr=expr,
            features=features,
            target=target,
            cache=cache,
            dest_col=dest_col,
        )

    def set_params(self, **kwargs):
        """
        Create a new Step with updated parameters.

        Parameters
        ----------
        **kwargs
            Parameter names and values to update.

        Returns
        -------
        Step
            A new Step instance with updated parameters.

        Examples
        --------
        >>> knn_step = Step(typ=KNeighborsClassifier, name="knn")
        >>> updated_step = knn_step.set_params(n_neighbors=10, weights="distance")
        """
        return self.__class__.from_instance_name(
            self.instance.set_params(**kwargs),
            name=self.name,
        )

    @classmethod
    def from_instance_name(cls, instance, name=None, deep=False):
        """
        Create a Step from an existing scikit-learn estimator instance.

        Parameters
        ----------
        instance : object
            A scikit-learn estimator instance.
        name : str, optional
            Name for the step. If None, generates from instance class name.

        Returns
        -------
        Step
            A new Step wrapping the estimator instance.
        """
        params_tuple = tuple(instance.get_params(deep=deep).items())
        return cls(typ=instance.__class__, name=name, params_tuple=params_tuple)

    @classmethod
    def from_name_instance(cls, name, instance, deep=False):
        """
        Create a Step from a name and estimator instance.

        Parameters
        ----------
        name : str
            Name for the step.
        instance : object
            A scikit-learn estimator instance.

        Returns
        -------
        Step
            A new Step wrapping the estimator instance.
        """
        return cls.from_instance_name(instance, name, deep=deep)

    @classmethod
    def from_fit_transform(
        cls, fit, transform, return_type, klass_name=None, name=None
    ):
        """
        Create a Step from custom fit and transform functions.

        Parameters
        ----------
        fit : callable
            Function to fit the model.
        transform : callable
            Function to transform with.
        return_type : DataType
            The return type for the transformation.
        klass_name : str, optional
            Name for the generated estimator class.
        name : str, optional
            Name for the step.

        Returns
        -------
        Step
            A new Step with a dynamically created transform type.
        """

        typ = make_estimator_typ(
            fit=fit, transform=transform, return_type=return_type, name=klass_name
        )
        return cls(typ=typ, name=name)

    @classmethod
    def from_fit_predict(cls, fit, predict, return_type, klass_name=None, name=None):
        """
        Create a Step from custom fit and predict functions.

        Parameters
        ----------
        fit : callable
            Function to fit the model.
        predict : callable
            Function to make predictions.
        return_type : DataType
            The return type for predictions.
        klass_name : str, optional
            Name for the generated estimator class.
        name : str, optional
            Name for the step.

        Returns
        -------
        Step
            A new Step with a dynamically created estimator type.
        """

        typ = make_estimator_typ(
            fit=fit, predict=predict, return_type=return_type, name=klass_name
        )
        return cls(typ=typ, name=name)

    __dask_tokenize__ = normalize_attrs


@frozen
class FittedStep:
    step = field(validator=instance_of(Step))
    expr = field(validator=instance_of(Expr))
    features = field(
        validator=optional(deep_iterable(instance_of(str), instance_of(tuple))),
        default=None,
        converter=tuple,
    )
    target = field(validator=optional(instance_of(str)), default=None)
    cache = field(
        validator=optional(
            instance_of(
                (
                    ParquetCache,
                    SourceCache,
                    ParquetSnapshotCache,
                    ParquetTTLSnapshotCache,
                )
            )
        ),
        default=None,
    )
    dest_col = field(validator=optional(instance_of(str)), default=None)

    def __attrs_post_init__(self):
        # we must have at least transform or predict
        if not (self.is_transform or self.is_predict):
            raise ValueError("Step must have transform or predict method")
        # if we are predict-only, we must have target (except for clustering)
        if self.target is None and self.is_predict and not self.is_transform:
            from sklearn.base import ClusterMixin

            if not isinstance(self.instance, ClusterMixin):
                raise ValueError(
                    f"Predict-only estimator {self.step.typ.__name__} requires a target column. "
                    "Pass target='column_name' to fit()."
                )
        # we can do very simple feature inference
        if self.features is None:
            features = tuple(col for col in self.expr.columns if col != self.target)
            object.__setattr__(self, "features", features)
        # we should now have everything fixed

    @property
    def instance(self):
        return self.step.instance

    @property
    def is_transform(self):
        return hasattr(self.step.typ, "transform")

    @property
    def is_predict(self):
        return hasattr(self.step.typ, "predict")

    @property
    def predict_return_type(self):
        return get_predict_return_type(self)

    @property
    @functools.cache
    def _deferred_fit_transform(self):
        assert self.is_transform
        return DeferredFitOther.from_fitted_step(self, mode="transform")

    @property
    @functools.cache
    def _deferred_fit_predict(self):
        assert self.is_predict
        return DeferredFitOther.from_fitted_step(self, mode="predict")

    @property
    @functools.cache
    def _deferred_fit_other(self):
        # Backward compat: prefer transform, fall back to predict
        return (
            self._deferred_fit_transform
            if self.is_transform
            else self._deferred_fit_predict
        )

    @property
    def deferred_model(self):
        return self._deferred_fit_other.deferred_model

    @property
    def model_udf(self):
        return self._deferred_fit_other.deferred_other

    @property
    def deferred_transform(self):
        return (
            self._deferred_fit_transform.deferred_other if self.is_transform else None
        )

    @property
    def deferred_predict(self):
        return self._deferred_fit_predict.deferred_other if self.is_predict else None

    @property
    def deferred_predict_proba(self):
        (attrname, fn) = (ResponseMethod.PREDICT_PROBA, predict_proba_sklearn)
        return (
            self._deferred_fit_other.make_deferred_other(
                fn=fn,
                return_type=dt.Array(dt.float64),
                name_infix=attrname.rstrip("_"),
            )
            if hasattr(self.step.instance.__class__, attrname)
            else None
        )

    @property
    def deferred_decision_function(self):
        (attrname, fn) = (ResponseMethod.DECISION_FUNCTION, decision_function_sklearn)
        return (
            self._deferred_fit_other.make_deferred_other(
                fn=fn,
                return_type=dt.Array(dt.float64),
                name_infix=attrname.rstrip("_"),
            )
            if hasattr(self.step.instance.__class__, attrname)
            else None
        )

    @property
    def deferred_feature_importances(self):
        (attrname, fn) = ("feature_importances_", feature_importances_sklearn)
        return (
            self._deferred_fit_other.make_deferred_other(
                fn=fn,
                return_type=dt.Array(dt.float64),
                name_infix=attrname.rstrip("_"),
            )
            if hasattr(self.step.instance.__class__, attrname)
            else None
        )

    @property
    @functools.cache
    def model(self):
        import pandas as pd

        match obj := self.deferred_model.execute():
            case pd.DataFrame():
                ((obj,),) = obj.values
            case bytes():
                pass
            case _:
                raise ValueError(f"unexpected deferred model result type {type(obj)}")
        return pickle.loads(obj)

    @property
    @functools.cache
    @cexcepts(ValueError)
    def structer(self):
        return Structer.from_instance_expr(
            self.step.instance, self.expr, features=self.features
        )

    def get_others(self, expr):
        others = tuple(other for other in expr.columns if other not in self.features)
        return others

    def transform_unpack(self, expr, retain_others=True, name="to_unpack"):
        struct_col = self.transform_raw(expr).name(name)
        if retain_others and (others := self.get_others(expr)):
            expr = expr.select(*others, struct_col)
        else:
            expr = struct_col.as_table()
        return expr.unpack(name)

    def transform_raw(self, expr, name=None):
        # when you use expr.mutate, you want transform_raw
        transformed = self.deferred_transform.on_expr(expr).name(
            name or self.dest_col or "transformed"
        )
        return transformed

    @property
    def tag_kwargs(self):
        return {
            "tag": "FittedStep-transform"
            if self.is_transform
            else "FittedStep-predict",
            **self.step.tag_kwargs,
            "features": self.features,
        }

    def transform(self, expr, retain_others=True):
        col = self.transform_raw(expr)
        others = self.get_others(expr) if retain_others else ()
        expr = expr.select(*others, col) if others else col.as_table()
        return self.structer.maybe_unpack(expr, col.get_name()).tag(
            **self.tag_kwargs,
        )

    @property
    def transformed(self):
        return self.transform(self.expr)

    def predict_raw(self, expr, name=None):
        col = self.deferred_predict.on_expr(expr).name(
            name or self.dest_col or ResponseMethod.PREDICT.value
        )
        return col

    def predict(self, expr, retain_others=True, name=None):
        col = self.predict_raw(expr, name=name)
        if retain_others and (others := self.get_others(expr)):
            expr = expr.select(*others, col)
        else:
            expr = col.as_table()
        return expr.tag(
            **self.tag_kwargs,
        )

    def mutate(self, expr, name=None):
        if self.is_predict:
            return self.predict_raw(expr, name=name)
        elif self.is_transform:
            return self.transform_raw(expr, name=name)
        else:
            raise ValueError("step has neither predict nor transform method")

    @property
    def predicted(self):
        return self.predict(self.expr)

    @toolz.curry
    def invoke_method_raw(self, expr, name=None, *, methodname):
        if not (method := getattr(self, f"deferred_{methodname}", None)):
            raise AttributeError(
                f"'{self.step.typ.__name__}' object has no attribute '{methodname}'"
            )
        return method.on_expr(expr).name(
            name or getattr(methodname, "value", methodname)
        )

    @toolz.curry
    def invoke_method(self, expr, retain_others=True, name=None, *, methodname):
        col = self.invoke_method_raw(expr=expr, name=name, methodname=methodname)
        expr = (
            expr.select(*others, col)
            if retain_others and (others := self.get_others(expr))
            else col.as_table()
        )
        return expr.tag(**self.tag_kwargs)

    predict_proba_raw = invoke_method_raw(methodname=ResponseMethod.PREDICT_PROBA)

    predict_proba = invoke_method(methodname=ResponseMethod.PREDICT_PROBA)

    decision_function_raw = invoke_method_raw(
        methodname=ResponseMethod.DECISION_FUNCTION
    )

    decision_function = invoke_method(methodname=ResponseMethod.DECISION_FUNCTION)

    feature_importances_raw = invoke_method_raw(methodname="feature_importances")

    def feature_importances(self, expr=None, name=None):
        import xorq.api as xo

        schema = self.expr.select(self.features).schema()
        empty_table = xo.memtable([dict.fromkeys(schema.names)], schema=schema)
        col = self.feature_importances_raw(empty_table, name)
        return col.as_table().tag(**self.tag_kwargs)


@frozen
class Pipeline:
    """
    A machine learning pipeline that chains multiple processing steps together.

    This class provides a xorq-native implementation that wraps scikit-learn pipelines,
    enabling deferred execution and integration with xorq expressions. The pipeline
    can contain both transform steps (data preprocessing) and a final prediction step.

    Parameters
    ----------
    steps : tuple of Step
        Sequence of Step objects that make up the pipeline.

    Attributes
    ----------
    steps : tuple of Step
        The sequence of processing steps.
    instance : sklearn.pipeline.Pipeline
        The equivalent scikit-learn Pipeline instance.
    transform_steps : tuple of Step
        All steps except the final prediction step (if any).
    predict_step : Step or None
        The final step if it has a predict method, otherwise None.

    Examples
    --------
    Create a pipeline from scikit-learn estimators:

    >>> from xorq.ml import Pipeline
    >>> from sklearn.preprocessing import StandardScaler
    >>> from sklearn.neighbors import KNeighborsClassifier
    >>> import sklearn.pipeline
    >>>
    >>> sklearn_pipeline = sklearn.pipeline.Pipeline([
    ...     ("scaler", StandardScaler()),
    ...     ("knn", KNeighborsClassifier(n_neighbors=5))
    ... ])
    >>> xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)

    Fit and predict with xorq expressions:

    >>> # Assuming train and test are xorq expressions
    >>> fitted = xorq_pipeline.fit(train, features=("feature1", "feature2"), target="target")  # quartodoc: +SKIP
    >>> predictions = fitted.predict(test)  # quartodoc: +SKIP

    Update pipeline parameters:

    >>> updated_pipeline = xorq_pipeline.set_params(knn__n_neighbors=10)  # quartodoc: +SKIP

    Notes
    -----
    - The Pipeline class is frozen (immutable) using attrs.
    - Pipelines automatically detect transform vs predict steps based on method availability.
    - The fit() method returns a FittedPipeline that can transform and predict on new data.
    - Parameter updates use sklearn's parameter naming convention (step__parameter).
    """

    steps = field(
        validator=deep_iterable(instance_of(Step), instance_of(tuple)),
        converter=tuple,
    )

    def __attrs_post_init__(self):
        assert self.steps

    @property
    def instance(self):
        """
        Create an equivalent scikit-learn Pipeline instance.

        Returns
        -------
        sklearn.pipeline.Pipeline
            A scikit-learn pipeline with the same steps and parameters.
        """

        import sklearn

        return sklearn.pipeline.Pipeline(
            tuple((step.name, step.instance) for step in self.steps)
        )

    @property
    def transform_steps(self):
        """
        Get all transformation steps (excluding final prediction step).

        Returns
        -------
        tuple of Step
            All steps that transform data but don't make final predictions.
        """

        (*steps, last_step) = self.steps
        if hasattr(last_step.instance, "predict"):
            return steps
        else:
            return self.steps

    @property
    def predict_step(self):
        """
        Get the final prediction step if it exists.

        Returns
        -------
        Step or None
            The final step if it has a predict method, otherwise None.
        """

        (*_, last_step) = self.steps
        return last_step if hasattr(last_step.instance, "predict") else None

    def fit(self, expr, features=None, target=None, cache=None):
        """
        Fit the pipeline to training data.

        This method sequentially fits each step in the pipeline, using the output
        of each transform step as input to the next step.

        Parameters
        ----------
        expr : Expr
            The xorq expression containing training data.
        features : tuple of str, optional
            Column names to use as features. If None, infers from expr columns
            excluding the target.
        target : str, optional
            Target column name. Required if pipeline has a prediction step.
        cache : Cache, optional
            Storage backend for caching fitted models.

        Returns
        -------
        FittedPipeline
            A fitted pipeline that can transform and predict on new data.

        Raises
        ------
        ValueError
            If target is not provided but pipeline has a prediction step.

        Examples
        --------
        >>> fitted = pipeline.fit(
        ...     train_data,
        ...     features=("sepal_length", "sepal_width"),
        ...     target="species"
        ... )  # quartodoc: +SKIP
        """

        if not target and self.predict_step:
            raise ValueError("Can't infer target for a prediction step")
        features = features or tuple(col for col in expr.columns if col != target)
        fitted_steps = ()
        transformed = expr
        # During fit, other (non-feature) columns are only needed if a predict
        # step will require the target column in the intermediate result.
        retain_others_during_fit = bool(self.predict_step)
        for step in self.transform_steps:
            fitted_step = step.fit(
                transformed,
                features=features,
                target=target,
                cache=cache,
            )
            fitted_steps += (fitted_step,)
            transformed = fitted_step.transform(
                transformed, retain_others=retain_others_during_fit
            )
            # KV-encoded outputs (e.g. OneHotEncoder, SelectKBest) flow through
            # struct field extraction in nested DataFusion queries; materialization
            # avoids type-propagation failures for list<struct> columns.
            if fitted_step.structer.has_kv_output:
                transformed = transformed.pipe(do_into_backend)
            features = fitted_step.structer.get_output_columns(
                fitted_step.dest_col or "transformed"
            )
        if step := self.predict_step:
            fitted_step = step.fit(
                transformed,
                features=features,
                target=target,
                cache=cache,
            )
            fitted_steps += (fitted_step,)
            # transformed = fitted_step.transform(transformed)
        return FittedPipeline(fitted_steps, expr)

    @classmethod
    def from_instance(cls, instance, deep=False):
        """
        Create a Pipeline from an existing scikit-learn Pipeline.

        Parameters
        ----------
        instance : sklearn.pipeline.Pipeline
            A fitted or unfitted scikit-learn pipeline.

        Returns
        -------
        Pipeline
            A new xorq Pipeline wrapping the scikit-learn pipeline.

        Examples
        --------
        >>> import sklearn.pipeline
        >>> from sklearn.preprocessing import StandardScaler
        >>> from sklearn.svm import SVC
        >>>
        >>> sklearn_pipe = sklearn.pipeline.Pipeline([
        ...     ("scaler", StandardScaler()),
        ...     ("svc", SVC())
        ... ])
        >>> xorq_pipe = Pipeline.from_instance(sklearn_pipe)
        """
        # https://github.com/scikit-learn/scikit-learn/issues/18272#issuecomment-682180783
        steps = tuple(
            Step.from_instance_name(step, name, deep=deep)
            for name, step in instance.steps
        )
        return cls(steps)

    def set_params(self, **kwargs):
        return self.__class__.from_instance(self.instance.set_params(**kwargs))


@frozen
class FittedPipeline:
    fitted_steps = field(
        validator=deep_iterable(instance_of(FittedStep), instance_of(tuple)),
        converter=tuple,
    )
    expr = field(validator=instance_of(Expr))

    def __attrs_post_init__(self):
        assert self.fitted_steps

    @property
    def is_predict(self):
        (*_, last_step) = self.fitted_steps
        return hasattr(last_step.step.instance, "predict")

    @property
    def transform_steps(self):
        return self.fitted_steps[:-1] if self.is_predict else self.fitted_steps

    @property
    def predict_step(self):
        return self.fitted_steps[-1] if self.is_predict else None

    @property
    def tag_kwargs(self):
        return {
            "transforms_tags": tuple(
                tuple(fitted_step.tag_kwargs.items())
                for fitted_step in self.transform_steps
            ),
        }

    def transform(self, expr, tag=True):
        transformed = expr
        for fitted_step in self.transform_steps:
            transformed = fitted_step.transform(transformed)
            if fitted_step.structer.has_kv_output:
                transformed = transformed.pipe(do_into_backend)
        if tag:
            transformed = transformed.tag(
                "FittedPipeline-transform",
                **self.tag_kwargs,
            )
        return transformed

    def predict(self, expr, name=None):
        transformed = self.transform(expr, tag=False)
        return self.predict_step.predict(transformed, name=name).tag(
            "FittedPipeline-predict",
            predict_tags=tuple(self.predict_step.tag_kwargs.items()),
        )

    @toolz.curry
    def invoke_predict_method(self, expr, tag_name, tag_key, *, methodname, name=None):
        if not self.is_predict:
            raise ValueError("Pipeline does not have a predict step")
        if not (method := getattr(self.predict_step, methodname, None)):
            raise ValueError(f"predict step does not have a method named {methodname}")
        transformed = self.transform(expr, tag=False)
        predicted = method(transformed, name=name)
        # Array-typed UDF outputs (e.g. predict_proba returning Array[float64]) cause
        # issues in DataFusion when referenced multiple times in metric computations;
        # materialization avoids stream exhaustion on second reference.
        deferred_method = getattr(self.predict_step, f"deferred_{methodname}", None)
        if deferred_method is not None and isinstance(
            deferred_method.return_type, dt.Array
        ):
            predicted = predicted.pipe(do_into_backend)
        return predicted.tag(
            tag_name,
            **{tag_key: tuple(self.predict_step.tag_kwargs.items())},
        )

    def predict_proba(self, expr, name=None):
        return self.invoke_predict_method(
            expr,
            tag_name="FittedPipeline-predict_proba",
            tag_key="predict_proba_tags",
            methodname=ResponseMethod.PREDICT_PROBA,
            name=name,
        )

    def decision_function(self, expr, name=None):
        return self.invoke_predict_method(
            expr,
            tag_name="FittedPipeline-decision_function",
            tag_key="decision_function_tags",
            methodname=ResponseMethod.DECISION_FUNCTION,
            name=name,
        )

    def feature_importances(self, expr, name=None):
        return self.invoke_predict_method(
            expr,
            tag_name="FittedPipeline-feature_importances",
            tag_key="feature_importances_tags",
            methodname="feature_importances",
            name=name,
        )

    def _get_default_scorer(self):
        """Get the default scorer based on model type.

        Returns
        -------
        scorer
            An sklearn scorer object appropriate for the model type.

        Raises
        ------
        ValueError
            If model type is not recognized.
        """
        from xorq.expr.ml.metrics import _default_scorer_for_model

        return _default_scorer_for_model(self.predict_step.instance)

    def score_expr(self, expr, scorer=None, **kwargs):
        """Compute metrics using deferred execution.

        Parameters
        ----------
        expr : ibis.Expr
            Expression containing test data
        scorer : str, callable, _BaseScorer, Scorer, or None
            Scorer specification. If None, uses model's default.
            Automatically detects whether scorer needs predict, predict_proba,
            or decision_function.
        **kwargs : dict
            Additional arguments passed to the metric function

        Returns
        -------
        ibis.Expr
            Deferred metric expression
        """
        from xorq.expr.ml.metrics import Scorer, deferred_sklearn_metric

        if scorer is None:
            s = Scorer.from_model(self.predict_step.instance)
        else:
            s = Scorer.from_spec(scorer)

        # Route predictions based on response_method
        method = ResponseMethod(s.response_method)
        expr_with_preds = getattr(self, method.value)(expr)
        pred = method.value

        return deferred_sklearn_metric(
            expr=expr_with_preds,
            target=self.predict_step.target,
            pred=pred,
            metric=s,
            metric_kwargs=kwargs,
        )

    def score(self, X, y, scorer=None, **kwargs):
        """Compute model score on test data.

        Parameters
        ----------
        X : array-like
            Test features
        y : array-like
            Test targets
        scorer : str or callable, optional
            Scorer name from sklearn.metrics.get_scorer_names() or a callable metric function.
            If None, uses model's default (accuracy for classifiers, r2 for regressors)
        **kwargs : dict
            Additional arguments passed to the scorer function

        Returns
        -------
        float
            The computed score
        """
        import numpy as np
        import pandas as pd

        from xorq.expr import api

        if not self.is_predict:
            raise ValueError("Pipeline does not have a predict step")

        # Use the first step's features (raw input names) so the full
        # pipeline (transform → predict) runs correctly.
        features = self.fitted_steps[0].features
        target = self.predict_step.target
        df = pd.DataFrame(
            np.array(X),
            columns=features,
        ).assign(**{target: y})
        expr = api.register(df, "t")
        return self.score_expr(expr, scorer=scorer, **kwargs).execute()


def get_target_type(step_instance, step, expr, features, target):
    return expr[target].type()


registry = Dispatch()


def get_predict_return_type(fitted_step):
    instance = fitted_step.step.instance
    return getattr(instance, "return_type", None) or registry(
        instance,
        fitted_step.step,
        fitted_step.expr,
        fitted_step.features,
        fitted_step.target,
    )


@registry.register(object)
def raise_on_unregistered(instance, step, expr, features, target):
    raise ValueError(f"Can't handle {instance.__class__.__name__}")


@registry.register_lazy("sklearn")
def lazy_register_sklearn():
    from sklearn.base import (
        ClassifierMixin,
        ClusterMixin,
        RegressorMixin,
    )
    from sklearn.ensemble import (
        RandomForestClassifier,
        RandomForestRegressor,
    )
    from sklearn.linear_model import (
        LinearRegression,
        LogisticRegression,
    )
    from sklearn.neighbors import (
        KNeighborsClassifier,
    )

    registry.register(LinearRegression, return_constant(dt.float))
    registry.register(LogisticRegression, get_target_type)
    registry.register(KNeighborsClassifier, get_target_type)
    registry.register(ClassifierMixin, get_target_type)
    registry.register(RandomForestRegressor, return_constant(dt.float))
    registry.register(RandomForestClassifier, get_target_type)
    registry.register(
        RegressorMixin, return_constant(dt.float)
    )  # General fallback for regressors
    registry.register(
        ClusterMixin, return_constant(dt.int64)
    )  # Clustering predict returns integer labels


get_predict_return_type.register = registry.register
